from collections import OrderedDict

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from tqdm import tqdm
from ..models import build_model
from . import BaseFlow, register_flow
from ..tasks import build_task


@register_flow("MeiREC_trainer")
class MeiRECTrainer(BaseFlow):

    def __init__(self, args):
        super().__init__(args)
        self.args = args
        self.model_name = args.model
        self.device = args.device
        self.task = build_task(args)

        self.model = build_model(self.model_name).build_model_from_args(
            self.args).model
        print("build_model_finish")
        self.model = self.model.to(self.device)
        self.loss_fn = nn.BCELoss(reduction='mean')
        # self.evaluator = self.task.evaluate
        self.optimizer = optim.Adam(self.model.parameters(), lr=args.lr)
        self.scheduler = optim.lr_scheduler.ExponentialLR(self.optimizer, 0.99)
        self.patience = args.patience
        self.max_epoch = args.max_epoch

    def train(self):
        train_loader = self.task.train_loader
        test_loader = self.task.test_loader

        for epoch in range(1, self.max_epoch + 1):
            pbar = tqdm(train_loader)
            if epoch % 3 == 0:
                for i, data in enumerate(pbar):
                    # train step, data: batch * features
                    if i != len(train_loader) - 1:
                        output, train_loss, train_auc, train_acc = self._train_step(data)
                        # record bar
                        pbar.set_description(f"epoch[{epoch}/{self.max_epoch}],\
                                data[{i}/{len(train_loader)}]")
                        pbar.set_postfix(
                            OrderedDict({
                                "train_loss": train_loss.item(),
                                "train_auc": train_auc,
                                "train_acc": train_acc
                            }))

                    else:
                        pr = []
                        la = []
                        _, train_loss, train_auc, train_acc = self._train_step(data)
                        for j, data_test in enumerate(test_loader):

                            output, test_loss, test_auc, test_acc = self.eval_model(data_test)
                            pr.append(output)
                            la.append(data_test["labels"])

                        prediction_test = torch.cat(pr, 0).cuda()
                        labels_test = torch.cat(la, 0).cuda()
                        loss_cross = self.loss_fn(prediction_test, labels_test.float())
                        loss_total = loss_cross + self.model.regular_loss
                        # auc and acc value
                        predicts_cpu = prediction_test.cpu().detach().numpy()
                        labels_cpu = labels_test.cpu().detach().numpy()
                        correct_num = sum((predicts_cpu > 0.5) == labels_cpu)
                        acc = correct_num / len(predicts_cpu)

                        # auc = roc_auc_score(labels_cpu, predicts_cpu)
                        auc = self.auc(labels_cpu, predicts_cpu)

                        pbar.set_description(f"epoch[{epoch}/{self.max_epoch}],\
                                                data[{i}/{len(train_loader)}]")
                        pbar.set_postfix(
                            OrderedDict({
                                "train_loss": train_loss.item(),
                                "train_auc": train_auc,
                                "train_acc": train_acc,
                                "test_loss": loss_total.item(),
                                "test_auc": auc,
                                "test_acc": acc
                            }))
            else:
                for i, data in enumerate(pbar):
                    # train step, data: batch * features
                    output, train_loss, train_auc, train_acc = self._train_step(data)
                    # record bar
                    pbar.set_description(f"epoch[{epoch}/{self.max_epoch}],\
                            data[{i}/{len(train_loader)}]")
                    pbar.set_postfix(
                        OrderedDict({
                            "train_loss": train_loss.item(),
                            "train_auc": train_auc,
                            "train_acc": train_acc
                        }))

            self.scheduler.step()

            # if epoch % 10 == 0:
            #     self.save_model(str(epoch))

            # self.save_model("latest.pth.tar")

    def _train_step(self, x):
        self.model.train()         # Sets the module in training mode
        output, loss, auc, acc = self.forward(x)
        self.optimizer.zero_grad()
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=5.0)
        self.optimizer.step()
        return output, loss, auc, acc

    def eval_model(self, x):
        self.model.eval()         # Sets the module in evaluation mode.

        with torch.no_grad():
            output, loss, auc, acc = self.forward(x)

        return output, loss, auc, acc

    def _confusion_matrix_at_thresholds(self, labels, preds, thresholds):

        values = {}
        tp = []
        fp = []
        tn = []
        fn = []

        for thr in thresholds:
            tp.append(sum((preds > thr) & (labels == 1)))
            fp.append(sum((preds > thr) & (labels == 0)))
            tn.append(sum((preds < thr) & (labels == 0)))
            fn.append(sum((preds < thr) & (labels == 1)))

        values['tp'] = torch.from_numpy(np.array(tp))
        values['fp'] = torch.from_numpy(np.array(fp))
        values['tn'] = torch.from_numpy(np.array(tn))
        values['fn'] = torch.from_numpy(np.array(fn))

        return values

    def auc(self, labels, preds, num_thresholds=200):   # tensorflow1.0_auc
        kepsilon = 1e-7                # to account for floating point imprecisions
        thresholds = [(i + 1) * 1.0 / (num_thresholds - 1)
                      for i in range(num_thresholds - 2)]
        thresholds = [0.0 - kepsilon] + thresholds + [1.0 + kepsilon]

        values = self._confusion_matrix_at_thresholds(
            labels, preds, thresholds)

        # Add epsilons to avoid dividing by 0.
        epsilon = 1.0e-6

        def compute_auc(tp, fn, tn, fp):

            rec = torch.div(tp + epsilon, tp + fn + epsilon)

            fp_rate = torch.div(fp, fp + tn + epsilon)
            x = fp_rate
            y = rec

            return torch.sum(
                torch.multiply(x[:num_thresholds - 1] - x[1:],
                          (y[:num_thresholds - 1] + y[1:]) / 2.))

        # sum up the areas of all the trapeziums
        auc_value = compute_auc(
            values['tp'], values['fn'], values['tn'], values['fp'])

        return auc_value.item()

    def forward(self, x):
        inputs = x['data'].T
        labels = x['labels']

        inputs = inputs.cuda()
        labels = labels.cuda()

        predicts = self.model(inputs)
        # loss
        loss_cross = self.loss_fn(predicts, labels.float())
        loss_total = loss_cross + self.model.regular_loss
        # auc and acc value
        predicts_cpu = predicts.cpu().detach().numpy()
        labels_cpu = labels.cpu().detach().numpy()
        correct_num = sum((predicts_cpu > 0.5) == labels_cpu)
        acc = correct_num / len(predicts_cpu)

        auc = self.auc(labels_cpu, predicts_cpu)

        return predicts, loss_total.cpu(), auc, acc




import dgl
import numpy as np
from openhgnn.models.MAGNN import mp_instance_sampler, mini_mp_instance_sampler

class MAGNN_sampler():
    '''
    Description
    -----------
    MAGNN sampler based on the implementation of author. The author only implements one sampling 1-layer subgraphs. Here
    we implement the MAGNN sampler which can sample n-layer subgraphs based on DGL and Pytorch.

    The MAGNN sampler samples n-layer subgraphs based on seed_node of a certain class, metapath instances of seed_nodes
    which contain all neighbors that might be sampled.

    Users can make use of the MAGNN sampler as a dataset of torch.DataLoader so that we can assign num_workers to
    concurrently speed up sampling subgraphs.

    Parameters
    ----------
    g : dgl.DGLGraph
        the dgl graph from which we sample subgraphs.
    mask : list or numpy.ndarray
        the train/test/test nodes mask, which should be like train mask = [False, True, False, True, False] indicating
        that the 1th, 4th nodes are training nodes of g. Note that the mask indicates which part of nodes in g will be
        sampled by MAGNN sampler. e.g. if it's the test mask, the sample will be employed on testing nodes of g to
        sample subgraphs for testing phase. This param is useful when the graph is too large to be directly validate on.
    num_layers : int
        the number of layers of each subgraph.
    category : any
        the class of seed_nodes.
    metapath_list : list
        the type of all metapaths in graph g. e.g. ['M-A-M', 'M-D-M', ...]
    num_samples : int
        the maximal number of metapath instances we'd like to sample for each node. i.e. for each node,
        if the number of metapath instances is more than num_samples, we sample num_samples instances randomly based on
        a specific calculated probability. Otherwise, we sample all instances.
    dataset_name : str
        the name of the dataset.
        Default : 'dblp4MAGNN'

    '''
    def __init__(self, g, mask, num_layers, category, metapath_list,
                 num_samples, dataset_name='dblp4MAGNN'):
        self.g = g
        self.mask = mask
        self.dataset_name = dataset_name
        self.metapath_list = metapath_list
        self.num_layers = num_layers
        self.category = category
        self.num_samples = num_samples
        self.mp_inst = mp_instance_sampler(g, self.metapath_list, self.dataset_name)

    def __getitem__(self, idx):
        '''
        Parameters
        ----------
        idx : int
            The node ids of ntype as self.category, that is, the original seed_nodes is {self.category: idx}.

        Returns
        -------
        dict
            The nids sampled by the MAGNN_sampler from the original graph. The nids are of original version.
        dict
            The metapath instances corresponding to the sampled graph. The nids in mp instances are of original version.
        DGLGraph
            The original heterogeneous subgraph.

        Notes
        -----
        Sampling neighbors based on seed_nodes and metapath instances mp_inst[i],
        we only need to sample mp_inst from all metapath instances with seed_nodes as mp_inst[i][0] and
        sample neighbors as mp_inst[i][1, 2, ...]. Here we consider mp_inst[i][0] as dst_nodes.

        But the sampled mp_inst above does not contain all the mp_inst of the sampled subgraph.
        For example, if seed_nodes contain M1 and we sample neighbors as A1, M2 based on metapath instance M1-A1-M2,
        we'd lose metapath instance like M2-A1-M1 because M2 is not one of the seed_nodes while M2-A1-M1 absolutely
        one of the metapath instances of the sampled subgraph.

        '''

        idx = np.where(self.mask)[0][idx]
        _seed_nodes = {self.category: idx}
        seed_nodes = {self.category: np.array([idx])}
        if self.num_layers < 1:
            raise ValueError("Wrong value of number of layers.")
        for _ in range(self.num_layers):
            mini_mp_inst = mini_mp_instance_sampler(seed_nodes=seed_nodes, mp_instances=self.mp_inst,
                                                    num_samples=self.num_samples)
            # seed_nodes = {}
            for metapath in mini_mp_inst.keys():
                _mini_mp_inst = mini_mp_inst[metapath]
                for i in range(0, len(metapath)):
                    if not metapath[i] in seed_nodes.keys():
                        seed_nodes[metapath[i]] = _mini_mp_inst[:, i]
                    else:
                        seed_nodes[metapath[i]] = np.concatenate((seed_nodes[metapath[i]], _mini_mp_inst[:, i]),
                                                                 axis=0)
            for ntype in seed_nodes.keys():
                seed_nodes[ntype] = np.unique(seed_nodes[ntype])

        for meta, idx in mini_mp_inst.items():
            mini_mp_inst[meta] = np.unique(np.concatenate((idx, np.flip(idx, axis=1))), axis=0)
        # Here seed_nodes are the nodes sampled from original graph with {self.category: idx} as _seed_nodes
        # while mini_mp_inst is corresponding mini metapath instances
        return seed_nodes, mini_mp_inst, _seed_nodes, self.g

    def __len__(self):
        # return self.g.number_of_nodes(self.category)
        return len(self.mask[self.mask == 1])

def collate_fn(batch):
    '''

    Parameters
    ----------
    batch : tuple
        The mini batch data, including seed_nodes, mini_mp_inst and g.
        The seed_nodes is the nids sampled by the MAGNN_sampler from the original graph.
        The mini_mp_inst is the metapath instances corresponding to the sampled graph.
        The g is the sampled graph.

    Returns
    -------
    tuple
        the sampled subgraph and corresponding mini metapath instances
    '''

    def convert_mp_nids(old_metapath_dict, old_nids):
        # convert the old_nids in metapath instances into the new_nids in the subgraph
        for meta in old_metapath_dict.keys():
            for i, ntype in enumerate(meta):
                old_metapath_dict[meta][:, i] = \
                    np.array(
                        list(map(lambda x: np.argwhere(old_nids[ntype] == x)[0][0], old_metapath_dict[meta][:, i]))
                    )

        return old_metapath_dict

    nids = {}
    mini_mp_inst = {}
    seed_nodes = {}

    ntypes = []
    meta_types = []
    seed_ntypes = []

    [ntypes.extend(list(_batch[0].keys())) for _batch in batch]
    [meta_types.extend(list(_batch[1].keys())) for _batch in batch]
    [seed_ntypes.extend(list(_batch[2].keys())) for _batch in batch]
    ntypes, meta_types, seed_ntypes = set(ntypes), set(meta_types), set(seed_ntypes)

    for _batch in batch:
        for ntype in ntypes:
            if ntype not in _batch[0].keys():
                continue
            if ntype in nids.keys():
                nids[ntype] = np.concatenate((nids[ntype], _batch[0][ntype]), axis=0)
            else:
                nids[ntype] = _batch[0][ntype]
        for meta_type in meta_types:
            if meta_type not in _batch[1].keys():
                continue
            if meta_type in mini_mp_inst.keys():
                mini_mp_inst[meta_type] = np.concatenate((mini_mp_inst[meta_type], _batch[1][meta_type]),
                                                         axis=0)
            else:
                mini_mp_inst[meta_type] = _batch[1][meta_type]
        for seed_ntype in seed_ntypes:
            if seed_ntype not in _batch[2].keys():
                continue
            _batch[2][seed_ntype] = np.array([_batch[2][seed_ntype]])
            if seed_ntype in seed_nodes.keys():
                seed_nodes[seed_ntype] = np.concatenate((seed_nodes[seed_ntype], _batch[2][seed_ntype]),
                                                        axis=0)
            else:
                seed_nodes[seed_ntype] = _batch[2][seed_ntype]

    for ntype in nids.keys():
        nids[ntype] = np.sort(np.unique(nids[ntype], axis=0), axis=0)
    for meta_type in mini_mp_inst.keys():
        mini_mp_inst[meta_type] = np.unique(mini_mp_inst[meta_type], axis=0)

    mini_mp_inst = convert_mp_nids(mini_mp_inst, nids)
    _subgraph = dgl.node_subgraph(batch[0][3], nids, store_ids=True)

    for seed_ntype in seed_ntypes:
        seed_nodes[seed_ntype] = np.unique(seed_nodes[seed_ntype], axis=0)
        _nids = _subgraph.nodes[seed_ntype].data[dgl.NID].cpu().numpy()
        seed_nodes[seed_ntype] = np.array(
            list(map(lambda x: np.argwhere(_nids == x)[0][0], seed_nodes[seed_ntype]))
        )
    # if seed_nodes = {'A':[0, 2]}, the 0th and 2th nodes of type 'A' are seed_nodes in
    # _subgraph
    return _subgraph, mini_mp_inst, seed_nodes



from ..trainerflow import build_flow
import optuna
from ..utils import set_random_seed


def hpo_experiment(args, trainerflow, **kwargs):
    logger = args.logger
    tool = AutoML(args, trainerflow, n_trials=args.hpo_trials, search_space=args.hpo_search_space, logger=logger)
    result = tool.run()
    logger.info("[Hyper-parameter optimization] Final results:{}".format(result))
    return result


class AutoML(object):
    """
    Args:
        search_space: function to obtain hyper-parameters to search
    """

    def __init__(self, args, trainerflow, n_trials=3, **kwargs):
        self.args = args
        self.trainerflow = trainerflow
        # self.seed = kwargs.pop("seed") if "seed" in kwargs else [1]
        assert "search_space" in kwargs
        self.search_space = kwargs["search_space"]
        self.metric = kwargs["metric"] if "metric" in kwargs else None
        self.logger = kwargs['logger']
        self.n_trials = n_trials
        self.best_score = None
        self.best_params = None
        self.default_params = kwargs

    def _objective(self, trials):
        args = self.args
        cur_params = self.search_space(trials)
        for key, value in cur_params.items():
            args.__setattr__(key, value)
        # Set random seed each time, or the initialization of the weight will be different.
        set_random_seed(args.seed)
        flow = build_flow(args, self.trainerflow)
        result = flow.train()['metric']['test']
        if isinstance(result, tuple):
            score = (result[0] + result[1]) / 2
        elif isinstance(result, dict):
            score = 0
            for _, v in result.items():
                score += v
            score /= len(result)
        if self.best_score is None or score > self.best_score:
            self.best_score = score
            self.best_params = cur_params
        return score

    def run(self):
        study = optuna.create_study(direction="maximize")
        study.optimize(self._objective, n_trials=self.n_trials, n_jobs=1)
        self.logger.info("[Hyper-parameter optimization] Best parameter: {}".format(self.best_params))
        return self.best_score



import dgl
import os
import random
import torch
from dgl.data.utils import load_graphs, download, extract_archive
import torch as th
from . import BaseDataset, register_dataset
from tqdm import tqdm



@register_dataset('abnorm_event_detection')
class AbnormEventDetectionDataset(BaseDataset):
    r"""
    The class *AbnormalEventDetectionDataset* is a class for datasets which can be used in task *abnormal event detection*.

    Attributes
    -------------
    g : dgl.DGLHeteroGraph
        The heterogeneous graph.
    in_dim : int
        Dimension of node features.
    event_features : tensor
        All event features.
    event_mask : tensor
        All event mask.
        It will be used to mask useless features.
    neg_event_features : tensor
        The neg event is event with fewer meta paths to target event.
    pos_event_features : tensor
        The pos event is event with more meta paths to target event.
    neg_context_features : tensor
        The neg context is that context of original event with partially replaced some nodes.
    neg_entity_features : tensor
        The neg entity is composed that node not with edge to target node.
    event_list : list
        The list composed event.
    pos_node_set_dict : dict
        The node's pos node set that node with edge to original node.
    neg_node_set_dict : dict
        The node's neg node set that node not with edge to original node.
    neg_event_set_list : list
        All event's neg event.
    pos_event_set_list : list
        All event's pos event.
    type_max_num : dict
        The max in the number of nodes of an event of a certain type.
    neg_num : int
        The size of neg entity.
    type_label_node_list : list
        All node of a type and a label
    context_type_num : dict
        A number used to represent a type.
    type_num : torch
        The type num will be used to compute embedding.
    max_type_features_len : int
        The max of dimension of all type node features.
    center_type : str
        The center node type from the event.
    context_type : list
        The context node type from the event.
    event_label : tensor
        The event label.
        0 indicates that it is not an abnormal event.
        1 indicates that it is an abnormal event.
    """

    def __init__(self, dataset_name, *args, **kwargs):
        super(AbnormEventDetectionDataset, self).__init__(*args, **kwargs)
        self.in_dim = None
        self.event_features = None
        self.event_mask = None
        self.neg_event_features = None
        self.pos_event_features = None
        self.neg_context_features = None
        self.neg_entity_features = None
        self.event_list = None
        self.pos_node_set_dict = None
        self.neg_node_set_dict = None
        self.neg_event_set_list = []
        self.pos_event_set_list = []
        self.type_max_num = None
        self.neg_num = 10
        self.type_label_node_list = None
        self.context_type_num = None
        self.type_num = None
        self.max_type_features_len = 0
        self.g, self.center_type, self.context_type, self.event_label = self.get_graph(dataset_name)
        print("get graph is Ok")
        self.preprocess()
        print("preprocess is Ok")
        self.get_complete_events_features()
        print("get features is Ok")

    def set_neg_num(self, neg_num):
        self.neg_num = neg_num

    def get_batch(self, batch_size, shuffle: bool = True, device='cpu'):
        """
        Get all batches dataset from the tensor dataset.

        Parameters
        ----------
        batch_size
            Size of each batch.
        shuffle
            True: use random shuffle.
            False: use original shuffle.
        device
            Use cpu or gpu to composed.

        Returns
        -------
            A list is composed all batch.
        """
        event_len = self.event_features[self.center_type].shape[0]
        shuffle_list = [n for n in range(event_len)]
        if shuffle:
            random.shuffle(shuffle_list)
        event_batch = self.get_batch_sample(batch_size, self.event_features, event_len, shuffle_list, device=device)
        neg_event_batch = self.get_batch_sample(batch_size, self.neg_event_features, event_len, shuffle_list,
                                                device=device)
        pos_event_batch = self.get_batch_sample(batch_size, self.pos_event_features, event_len, shuffle_list,
                                                device=device)
        neg_context_batch = self.get_batch_sample(batch_size, self.neg_context_features, event_len, shuffle_list,
                                                  device=device)
        neg_entity_batch = self.get_batch_sample(batch_size, self.neg_entity_features, event_len, shuffle_list,
                                                 device=device)
        event_mask = self.get_batch_mask(batch_size, self.event_mask, event_len, shuffle_list, device=device)

        type_num = self.get_batch_type_num(batch_size, self.type_num, event_len, shuffle_list, device=device)

        return event_batch, neg_event_batch, pos_event_batch, neg_context_batch, neg_entity_batch, event_mask, type_num

    def get_batch_type_num(self, batch_size, type_num: dict, event_len, shuffle_list, device='cpu'):
        type_num_batch_list = []
        type_num_batch_num = int(event_len / batch_size)
        type_num_ = dict()
        for key in type_num.keys():
            type_num_[key] = type_num[key][shuffle_list]
        for i in range(type_num_batch_num):
            type_num_batch = dict()
            for key in type_num.keys():
                type_num_batch[key] = type_num_[key][i * batch_size:(i + 1) * batch_size, :].to(device)
            type_num_batch_list.append(type_num_batch)
        if batch_size * type_num_batch_num < event_len:
            type_num_batch = dict()
            for key in type_num.keys():
                type_num_batch[key] = type_num_[key][type_num_batch_num * batch_size:, :].to(device)
            type_num_batch_list.append(type_num_batch)

        return type_num_batch_list


    def get_batch_mask(self, batch_size, mask: dict, event_len, shuffle_list, device='cpu'):
        mask_batch_list = []
        mask_batch_num = int(event_len / batch_size)
        mask_ = dict()
        for key in mask.keys():
            mask_[key] = mask[key][shuffle_list]
        for i in range(mask_batch_num):
            mask_batch = dict()
            for key in mask.keys():
                mask_batch[key] = mask_[key][i * batch_size:(i + 1) * batch_size, :].to(device)
            mask_batch_list.append(mask_batch)

        if batch_size * mask_batch_num < event_len:
            mask_batch = dict()
            for key in mask.keys():
                mask_batch[key] = mask_[key][mask_batch_num * batch_size:, :].to(device)
            mask_batch_list.append(mask_batch)

        return mask_batch_list

    def get_batch_sample(self, batch_size, event: dict, event_len, shuffle_list, device='cpu'):
        event_batch_list = []
        event_batch_num = int(event_len / batch_size)
        event_ = dict()
        for key in event.keys():
            event_[key] = event[key][shuffle_list]
        for i in range(event_batch_num):
            event_batch = dict()
            for key in event.keys():
                event_batch[key] = event_[key][i * batch_size:(i + 1) * batch_size, :, :].to(device)
            event_batch_list.append(event_batch)

        if batch_size * event_batch_num < event_len:
            event_batch = dict()
            for key in event.keys():
                event_batch[key] = event_[key][event_batch_num * batch_size:, :, :].to(device)
            event_batch_list.append(event_batch)

        return event_batch_list

    def get_graph(self, name_dataset):
        if name_dataset == 'aminer4AEHCL':
            data_path = './openhgnn/dataset/aminer4aehcl.zip'
            if not os.path.exists(data_path):
                # download file
                download(url='https://s3.cn-north-1.amazonaws.com.cn/dgl-data/dataset/aminer4aehcl.zip', path='./openhgnn/dataset/')
            extract_archive(data_path, './openhgnn/dataset/aminer4aehcl')
            g, label = load_graphs('./openhgnn/dataset/aminer4aehcl/aminer4aehcl.bin')
            g = g[0]
            event_label = label["event_label"]
            center = 'paper'
            context = ['conf', 'author']
        else:
            raise ValueError()
        return g, center, context, event_label

    def get_node_features(self, node):
        if len(node) == 3:
            return self.g.nodes[node[2]].data["features"][node[1]]
        return self.g.nodes[node[0]].data["features"][node[1]]

    def get_node_label(self, node):
        return int(self.g.nodes[node[0]].data["label"][node[1]])

    def preprocess(self):
        """
        Preprocess dataset.
        Changing heterogeneous graph dataset to that lists or sets are composed node.

        Returns
        -------
            None
        """
        self.context_type_num = dict()
        type_num = 0
        for tp in self.context_type:
            type_num += 1
            self.context_type_num[tp] = type_num

        self.event_list = []
        self.type_max_num = dict()
        events_context_type_node = []
        center_node_number = self.g.num_nodes(self.center_type)

        with tqdm(range(center_node_number), desc='get event') as tbar:
            for i in tbar:
                event = [(self.center_type, i)]
                edge_types_ = self.g.canonical_etypes
                event_context_type_node = dict()
                for edge_type_ in edge_types_:
                    edge_type = edge_type_[1]
                    node_type = edge_type_[2]
                    node_list = self.g.out_edges(i, etype=edge_type)[1]
                    # print(node_list)
                    event_context_type_node[node_type] = node_list.tolist()
                    for j in node_list:
                        event.append((node_type, int(j)))
                events_context_type_node.append(event_context_type_node)
                self.event_list.append(event)

        with tqdm(range(center_node_number), desc='get type num') as tbar:
            for i in tbar:
                event = self.event_list[i]
                type_num = dict()
                for node in event:
                    if node[0] not in type_num.keys():
                        type_num[node[0]] = 1
                    else:
                        type_num[node[0]] += 1
                for key in type_num.keys():
                    if key not in self.type_max_num.keys():
                        self.type_max_num[key] = type_num[key]
                    else:
                        if type_num[key] > self.type_max_num[key]:
                            self.type_max_num[key] = type_num[key]

        # get event's positive and negative event
        have_the_node_dict = dict()
        with tqdm(range(center_node_number), desc='process pos and neg event') as tbar:
            for i in tbar:
                for key in events_context_type_node[i].keys():
                    if key not in have_the_node_dict.keys():
                        have_the_node_dict[key] = dict()
                    for node_num in events_context_type_node[i][key]:
                        if node_num not in have_the_node_dict[key].keys():
                            have_the_node_dict[key][node_num] = []
                        have_the_node_dict[key][node_num].append(i)

        with tqdm(range(center_node_number), desc='get pos and neg event') as tbar:
            for i in tbar:
                pos_event_num_set = set()
                neg_event_num_set = set()
                meta_times = dict()
                for j in range(1000):
                    can_use = False
                    while can_use is False:
                        num = random.randint(0, center_node_number - 1)
                        can_use = True
                        for key in events_context_type_node[i].keys():
                            if len(set(events_context_type_node[i][key]) & set(events_context_type_node[num][key])) > 0:
                                can_use = False
                        if num in neg_event_num_set:
                            can_use = False
                    neg_event_num_set.add(num)

                for key in events_context_type_node[i].keys():
                    for node_num in events_context_type_node[i][key]:
                        for num in have_the_node_dict[key][node_num]:
                            if num != i:
                                if num not in meta_times.keys():
                                    meta_times[num] = 1
                                else:
                                    meta_times[num] = meta_times[num] + 1
                maxn = 0
                for key, value in meta_times.items():
                    if value > maxn:
                        maxn = value
                if maxn == 0:
                    for j in range(1000):
                        can_use = False
                        while can_use is False:
                            num = random.randint(0, center_node_number - 1)
                            can_use = True
                            if num in pos_event_num_set:
                                can_use = False
                        pos_event_num_set.add(num)
                else:
                    for key, value in meta_times.items():
                        if value == maxn:
                            pos_event_num_set.add(key)

                self.neg_event_set_list.append(neg_event_num_set)
                self.pos_event_set_list.append(pos_event_num_set)

        all_type_set = dict()
        relation_type = dict()
        self.pos_node_set_dict = dict()
        with tqdm(range(center_node_number), desc='get pos node') as tbar:
            for i in tbar:
                event = self.event_list[i]
                for node in event:
                    if node[0] not in all_type_set.keys():
                        all_type_set[node[0]] = set()
                    if node[0] not in self.pos_node_set_dict.keys():
                        self.pos_node_set_dict[node[0]] = dict()
                    if node[1] not in self.pos_node_set_dict[node[0]].keys():
                        self.pos_node_set_dict[node[0]][node[1]] = set()
                    all_type_set[node[0]].add(node)
                    for node1 in event:
                        if node != node1:
                            if node[0] not in relation_type.keys():
                                relation_type[node[0]] = set()
                            relation_type[node[0]].add(node1[0])
                            self.pos_node_set_dict[node[0]][node[1]].add(node1)

        self.neg_node_set_dict = dict()
        for key in all_type_set.keys():
            self.neg_node_set_dict[key] = dict()
            with tqdm(all_type_set[key], desc='get neg node '+key) as tbar:
                for node in tbar:
                    self.neg_node_set_dict[key][node[1]] = set()
                    for tp in relation_type[key]:
                        for node1 in all_type_set[tp]:
                            if node != node1 and (node1 not in self.pos_node_set_dict[node[0]][node[1]]):
                                self.neg_node_set_dict[key][node[1]].add(node1)
                    if len(self.neg_node_set_dict[key][node[1]]) > 1000:
                        temp = random.sample(self.neg_node_set_dict[key][node[1]], 1000)
                        del self.neg_node_set_dict[key][node[1]]
                        self.neg_node_set_dict[key][node[1]] = set(temp)

        self.max_type_features_len = self.g.nodes[self.center_type].data['features'].shape[1]
        for tp in self.context_type:
            if self.g.nodes[tp].data['features'].shape[1] > self.max_type_features_len:
                self.max_type_features_len = self.g.nodes[tp].data['features'].shape[1]

        self.type_label_node_list = dict()
        for key in self.context_type:
            self.type_label_node_list[key] = dict()
            key_size = self.g.num_nodes(key)
            for i in range(key_size):
                label = self.get_node_label((key, i))
                if label not in self.type_label_node_list[key].keys():
                    self.type_label_node_list[key][label] = []
                self.type_label_node_list[key][label].append((key, i))


    def get_complete_events_features(self):
        """
        Get tensor datasets from the node's sets or lists.

        Returns
        -------
            None
        """
        self_events = []
        neg_events = []
        pos_events = []
        neg_contexts = []
        neg_entities = []

        len_ = len(self.event_list)
        for i in range(len(self.event_list)):
            self_event = self.event_list[i]

            neg_event_num = random.sample(self.neg_event_set_list[i], 1)[0]
            neg_event = self.event_list[neg_event_num]

            pos_event_num = random.sample(self.pos_event_set_list[i], 1)[0]
            pos_event = self.event_list[pos_event_num]

            neg_entity = []
            for node in self_event:
                nd_list = random.sample(self.neg_node_set_dict[node[0]][node[1]], self.neg_num)
                ndd_list = []
                for nd in nd_list:
                    ndd_list.append(nd)
                neg_entity.append((node[0], ndd_list))

            neg_context = []
            type_context = dict()
            for node in self_event:
                if node[0] == self.center_type:
                    neg_context.append(node)
                else:
                    if node[0] not in type_context.keys():
                        type_context[node[0]] = []
                    type_context[node[0]].append(node)
            for key in type_context.keys():
                nd = random.sample(type_context[key], 1)[0]
                nd_label = self.get_node_label(nd)
                for nd_other in type_context[key]:
                    if nd_other != nd:
                        neg_context.append(nd_other)
                neg_nd_label = random.sample(self.type_label_node_list[key].keys(), 1)[0]
                neg_nd = random.sample(self.type_label_node_list[key][neg_nd_label], 1)[0]
                while neg_nd in type_context[key] or neg_nd_label == nd_label:
                    # print(key,": ","get neg node")
                    neg_nd_label = random.sample(self.type_label_node_list[key].keys(), 1)[0]
                    neg_nd = random.sample(self.type_label_node_list[key][neg_nd_label], 1)[0]
                neg_context.append(neg_nd)

            self_events.append(self_event)
            neg_events.append(neg_event)
            pos_events.append(pos_event)
            neg_contexts.append(neg_context)
            neg_entities.append(neg_entity)

        event_features, self.event_mask, self.type_num = self.get_events_features(self_events)
        in_dim = event_features[self.center_type].shape[-1]
        self.in_dim = in_dim
        self.event_features = event_features
        self.neg_event_features, _, _ = self.get_events_features(neg_events)
        self.pos_event_features, _, _ = self.get_events_features(pos_events)
        self.neg_context_features, _, _ = self.get_events_features(neg_contexts)
        self.neg_entity_features = self.get_entities_features(neg_entities)

    def get_entities_features(self, entities):
        entities_features = dict()
        for entity in entities:
            entity_features = dict()
            entity_features[self.center_type] = []
            for tp in self.context_type:
                entity_features[tp] = []
            for tp, node_list in entity:
                the_features = []
                for nd in node_list:
                    features = self.get_node_features(nd).tolist()
                    features += [0.] * (self.max_type_features_len - len(features))
                    the_features.append(features)
                entity_features[tp].append(the_features)
            for key in entity_features.keys():
                entity_features[key] += [[[0.] * self.max_type_features_len] * self.neg_num] * (
                            self.type_max_num[key] - len(entity_features[key]))
                if key not in entities_features.keys():
                    entities_features[key] = []
                entities_features[key].append(entity_features[key])
        for key in entities_features.keys():
            entities_features[key] = torch.tensor(entities_features[key])
        return entities_features

    def get_events_features(self, events):
        events_features = dict()
        masks = dict()
        type_nums = dict()
        for event in events:
            event_features = dict()
            event_features[self.center_type] = []
            mask = dict()
            mask[self.center_type] = []
            type_num = dict()
            for tp in self.context_type:
                event_features[tp] = []
                type_num[tp] = []
                mask[tp] = []
            for node in event:
                features = self.get_node_features(node).tolist()
                features += [0.] * (self.max_type_features_len - len(features))
                event_features[node[0]].append(features)
            for key in event_features:
                mask[key] = [1.] * len(event_features[key]) + [0.] * (self.type_max_num[key] - len(event_features[key]))
                if key != self.center_type:
                    type_num[key] = [self.context_type_num[key]] * len(event_features[key]) + [0] * (self.type_max_num[key] - len(event_features[key]))
                event_features[key] += [[0.] * self.max_type_features_len] * (
                            self.type_max_num[key] - len(event_features[key]))
                if key not in events_features.keys():
                    events_features[key] = []
                    masks[key] = []
                    if key != self.center_type:
                        type_nums[key] = []
                if key != self.center_type:
                    type_nums[key].append(type_num[key])
                events_features[key].append(event_features[key])
                masks[key].append(mask[key])
        for key in events_features.keys():
            events_features[key] = torch.FloatTensor(events_features[key])
            masks[key] = torch.FloatTensor(masks[key])
            if key != self.center_type:
                type_nums[key] = torch.LongTensor(type_nums[key])
        return events_features, masks, type_nums



import dgl
from . import register_model, BaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.nn import Parameter
import math
import dgl.function as fn


@register_model('GATNE-T')
class GATNE(BaseModel):
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(hg.num_nodes(), args.dim, args.edge_dim, hg.etypes, len(hg.etypes), args.att_dim)

    def __init__(
            self,
            num_nodes,
            embedding_size,
            embedding_u_size,
            edge_types,
            edge_type_count,
            att_dim,
    ):
        super(GATNE, self).__init__()
        self.num_nodes = num_nodes
        self.embedding_size = embedding_size
        self.embedding_u_size = embedding_u_size
        self.edge_types = edge_types
        self.edge_type_count = edge_type_count
        self.att_dim = att_dim

        self.node_embeddings = Parameter(torch.FloatTensor(num_nodes, embedding_size))
        self.node_type_embeddings = Parameter(
            torch.FloatTensor(num_nodes, edge_type_count, embedding_u_size)
        )
        self.trans_weights = Parameter(
            torch.FloatTensor(edge_type_count, embedding_u_size, embedding_size)
        )
        self.trans_weights_s1 = Parameter(
            torch.FloatTensor(edge_type_count, embedding_u_size, att_dim)
        )
        self.trans_weights_s2 = Parameter(torch.FloatTensor(edge_type_count, att_dim, 1))

        self.reset_parameters()

    def reset_parameters(self):
        self.node_embeddings.data.uniform_(-1.0, 1.0)
        self.node_type_embeddings.data.uniform_(-1.0, 1.0)
        self.trans_weights.data.normal_(std=1.0 / math.sqrt(self.embedding_size))
        self.trans_weights_s1.data.normal_(std=1.0 / math.sqrt(self.embedding_size))
        self.trans_weights_s2.data.normal_(std=1.0 / math.sqrt(self.embedding_size))

    # embs: [batch_size, embedding_size]
    def forward(self, block):
        input_nodes = block.srcdata[dgl.NID]
        output_nodes = block.dstdata[dgl.NID]
        batch_size = block.number_of_dst_nodes()
        node_embed = self.node_embeddings
        node_type_embed = []

        with block.local_scope():
            for i in range(self.edge_type_count):
                edge_type = self.edge_types[i]
                block.srcdata[edge_type] = self.node_type_embeddings[input_nodes, i]
                block.dstdata[edge_type] = self.node_type_embeddings[output_nodes, i]
                block.update_all(
                    fn.copy_u(edge_type, "m"), fn.sum("m", edge_type), etype=edge_type
                )
                node_type_embed.append(block.dstdata[edge_type])

            node_type_embed = torch.stack(node_type_embed, 1)
            tmp_node_type_embed = node_type_embed.unsqueeze(2).view(
                -1, 1, self.embedding_u_size
            )
            trans_w = (
                self.trans_weights.unsqueeze(0)
                    .repeat(batch_size, 1, 1, 1)
                    .view(-1, self.embedding_u_size, self.embedding_size)
            )
            trans_w_s1 = (
                self.trans_weights_s1.unsqueeze(0)
                    .repeat(batch_size, 1, 1, 1)
                    .view(-1, self.embedding_u_size, self.att_dim)
            )
            trans_w_s2 = (
                self.trans_weights_s2.unsqueeze(0)
                    .repeat(batch_size, 1, 1, 1)
                    .view(-1, self.att_dim, 1)
            )

            attention = (
                F.softmax(
                    torch.matmul(
                        torch.tanh(torch.matmul(tmp_node_type_embed, trans_w_s1)),
                        trans_w_s2,
                    )
                        .squeeze(2)
                        .view(-1, self.edge_type_count),
                    dim=1,
                )
                    .unsqueeze(1)
                    .repeat(1, self.edge_type_count, 1)
            )

            node_type_embed = torch.matmul(attention, node_type_embed).view(
                -1, 1, self.embedding_u_size
            )
            node_embed = node_embed[output_nodes].unsqueeze(1).repeat(
                1, self.edge_type_count, 1
            ) + torch.matmul(node_type_embed, trans_w).view(
                -1, self.edge_type_count, self.embedding_size
            )
            last_node_embed = F.normalize(node_embed, dim=2)

            return last_node_embed  # [batch_size, edge_type_count, embedding_size]


class NSLoss(nn.Module):
    def __init__(self, num_nodes, num_sampled, embedding_size):
        super(NSLoss, self).__init__()
        self.num_nodes = num_nodes
        self.num_sampled = num_sampled
        self.embedding_size = embedding_size
        self.weights = Parameter(torch.FloatTensor(num_nodes, embedding_size))
        # [ (log(i+2) - log(i+1)) / log(num_nodes + 1)]
        self.sample_weights = F.normalize(
            torch.Tensor(
                [
                    (math.log(k + 2) - math.log(k + 1)) / math.log(num_nodes + 1)
                    for k in range(num_nodes)
                ]
            ),
            dim=0,
        )

        self.reset_parameters()

    def reset_parameters(self):
        self.weights.data.normal_(std=1.0 / math.sqrt(self.embedding_size))

    def forward(self, input, embs, label):
        n = input.shape[0]
        log_target = torch.log(
            torch.sigmoid(torch.sum(torch.mul(embs, self.weights[label]), 1))
        )
        negs = torch.multinomial(
            self.sample_weights, self.num_sampled * n, replacement=True
        ).view(n, self.num_sampled)
        noise = torch.neg(self.weights[negs])
        sum_log_sampled = torch.sum(
            torch.log(torch.sigmoid(torch.bmm(noise, embs.unsqueeze(2)))), 1
        ).squeeze()

        loss = log_target + sum_log_sampled
        return -loss.sum() / n



import math

import torch
import torch as th
import torch.nn as nn
from openhgnn.models import BaseModel, register_model


@register_model("AEHCL")
class AEHCL(BaseModel):
    r"""
        **Title:** Abnormal Event Detection via Hypergraph Contrastive Learning

        **Authors:** Bo Yan, Cheng Yang, Chuan Shi, Jiawei Liu, Xiaochen Wang

        AEHCL was introduced in `[paper] <http://www.shichuan.org/doc/145.pdf>`_
        and parameters are defined as follows:

        Parameters
        ----------
        center_type: str
            Center node's type.
        context_type: list(str)
            All context node's type.
        in_dim: int
            Node features' dimension.
        out_dim: int
            Hidden layers' dimension.
        num_of_attention_heads: int
            The number of attention heads.
        batch_size: int
            The size of each batch.
        all_len: int
            The number of nodes of an event.
        weight_inter: float
            The weight of inter loss function.
        weight_intra_multi: float
            The weight of intra multi loss function.
        weight_intra_pair: float
            The weight of intra pair loss function.
        t: float
            Temperature.

    """

    @classmethod
    def build_model_from_args(cls, args):
        return AEHCL(center_type=args.center_type, context_type=args.context_type, in_dim=args.in_dim,
                     out_dim=args.hidden_dim, num_of_attention_heads=args.num_of_attention_heads,
                     batch_size=args.batch_size, all_len=args.all_len, weight_inter=args.weight_inter,
                     weight_intra_multi=args.weight_intra_multi, weight_intra_pair=args.weight_intra_pair, t=args.t)

    def __init__(self, center_type, context_type, in_dim, out_dim, num_of_attention_heads, batch_size, all_len,
                 weight_inter, weight_intra_multi, weight_intra_pair, t):
        super(AEHCL, self).__init__()

        self.weight_inter = weight_inter
        self.weight_intra_multi = weight_intra_multi
        self.weight_intra_pair = weight_intra_pair

        self.fc = nn.ModuleDict()
        self.fc[center_type] = nn.Sequential(
            nn.Linear(in_dim, out_dim, bias=True),
            nn.ELU(),
            nn.Linear(out_dim, out_dim, bias=True),
        )
        for tp in context_type:
            self.fc[tp] = nn.Sequential(
                nn.Linear(in_dim, out_dim, bias=True),
                nn.ELU(),
                nn.Linear(out_dim, out_dim, bias=True),
            )

        for key in self.fc.keys():
            for model in self.fc[key]:
                if isinstance(model, nn.Linear):
                    nn.init.xavier_normal_(model.weight, gain=1.414)

        self.fc3 = nn.Linear(out_dim * 3, out_dim, bias=False)

        self.ebl = nn.Bilinear(out_dim * 2, out_dim * 2, 1)
        self.ebl_t = nn.Bilinear(out_dim, out_dim, 1)

        self.selfatt = SelfAttention(center_type, context_type, out_dim, num_of_attention_heads, len(context_type) + 1)
        self.maxpool = nn.MaxPool2d(kernel_size=(all_len - 1, 1))
        self.att = Attention(center_type, context_type, out_dim)

        self.sigmoid = nn.Sigmoid()

        self.BCE_loss = nn.BCEWithLogitsLoss(reduction='none', pos_weight=torch.tensor([1.], dtype=torch.float32))

        self.center_type = center_type
        self.batch_size = batch_size
        self.t = t

    def forward(self, event, pos_event, neg_event, neg_context, neg_entity, event_mask, type_num):

        # inter

        # event
        event_center = event[self.center_type]  # (batch,1,in_dim)
        event_center_z = self.fc[self.center_type](event_center)  # (batch,1,out_dim)

        event_context = dict()
        event_context_z = dict()
        for key in event.keys():
            if key != self.center_type:
                event_context[key] = event[key]  # (batch,len[key],in_dim)
                event_context_z[key] = self.fc[key](event_context[key])  # (batch,len[key],out_dim)

        event_z = dict()
        event_z[self.center_type] = event_center_z
        for key in event_context_z.keys():
            event_z[key] = event_context_z[key]

        event_features = self.att(event_z)  # (batch,out_dim*2)

        # pos
        pos_event_z = dict()
        for key in pos_event.keys():
            pos_event_z[key] = self.fc[key](pos_event[key])

        pos_event_features = self.att(pos_event_z)  # (batch,out_dim*2)
        pos_inter = self.ebl(pos_event_features, event_features)  # (batch,1)

        # neg
        neg_event_z = dict()
        for key in neg_event.keys():
            neg_event_z[key] = self.fc[key](neg_event[key])

        neg_event_features = self.att(neg_event_z)  # (batch,out_dim*2)
        neg_inter = self.ebl(neg_event_features, event_features)  # (batch,1)

        # intra_multi

        # pos
        pos_context = None
        type_num_ = None
        for key in event_context_z.keys():
            if pos_context is None:
                pos_context = event_context_z[key]
                type_num_ = type_num[key]
            else:
                pos_context = torch.cat((pos_context, event_context_z[key]), 1)  # (batch,all_len-1,out_dim)
                type_num_ = torch.cat((type_num_, type_num[key]), 1) # (batch,all_len-1)
        pos_context = self.selfatt(pos_context, type_num_)  # (batch,all_len-1,out_dim)
        pos_context = self.maxpool(pos_context)  # (batch,1,out_dim)
        pos_multi = self.ebl_t(pos_context, event_center_z)  # (batch,1,1)
        pos_multi = torch.squeeze(pos_multi, -1)  # (batch,1)

        # neg
        neg_context_h = dict()
        for key in neg_context.keys():
            if key != self.center_type:
                neg_context_h[key] = self.fc[key](neg_context[key])
        neg_context_ = None
        for key in neg_context_h.keys():
            if neg_context_ is None:
                neg_context_ = neg_context_h[key]
            else:
                neg_context_ = torch.cat((neg_context_, neg_context_h[key]), 1)
        neg_context_ = self.selfatt(neg_context_, type_num_)
        neg_context_ = self.maxpool(neg_context_)
        neg_multi = self.ebl_t(neg_context_, event_center_z)  # (batch,1,1)
        neg_multi = torch.squeeze(neg_multi, -1)  # (batch,1)

        # intra_pair

        # pos
        pos_entity_z = event_center_z  # (batch,1,out_dim)
        pos_entity_mask = event_mask[self.center_type] # (batch,1)
        for key in event_context_z.keys():
            pos_entity_z = torch.cat((pos_entity_z, event_context_z[key]), 1)  # (batch,all_len,out_dim)
            pos_entity_mask = torch.cat((pos_entity_mask, event_mask[key]), -1) # (batch,all_len)
        pos_norm = torch.norm(pos_entity_z, dim=-1, keepdim=True)
        pos_entity_z = pos_entity_z / pos_norm  # (batch,all_len,out_dim)
        pos_matrix = torch.matmul(pos_entity_z, torch.transpose(pos_entity_z, 1, 2))  # (batch,all_len,all_len)
        pos_matrix = torch.exp(pos_matrix / self.t)  # (batch,all_len,all_len)
        mask = torch.ones_like(pos_matrix) - torch.eye(pos_matrix.shape[1]).to(pos_matrix.device)
        pos_matrix_ = torch.mul(pos_matrix, mask)  # (batch,all_len,all_len)
        pos_entity_mask_ = torch.unsqueeze(pos_entity_mask, 1) # (batch,1,all_len)
        pos_entity_mask_ = pos_entity_mask_.expand(pos_matrix_.shape) # (batch,all_len,all_len)
        pos_matrix_ = pos_matrix_ * pos_entity_mask_ # (batch,all_len,all_len)


        # neg
        neg_entity_z = None
        for key in neg_entity.keys():
            if neg_entity_z is None:
                neg_entity_z = self.fc[key](neg_entity[key])
            else:
                neg_entity_z = torch.cat((neg_entity_z, self.fc[key](neg_entity[key])), 1) # (batch,all_len,neg_num,out_dim)
        neg_norm = torch.norm(neg_entity_z, dim=-1, keepdim=True) # (batch,all_len,neg_num,1)
        neg_entity_z = neg_entity_z / neg_norm # (batch,all_len,neg_num,out_dim)
        neg_matrix = torch.matmul(torch.unsqueeze(pos_entity_z, 2), torch.transpose(neg_entity_z, 2, 3)) # (batch,all_len,1,neg_num)
        neg_matrix = torch.squeeze(neg_matrix) # (batch,all_len,neg_num)
        neg_matrix_ = torch.exp(neg_matrix / self.t)  # (batch,all_len,neg_num)

        # pos_inter, neg_inter, pos_multi, neg_multi, pos_matrix, neg_matrix -> lose / score
        lbl = torch.unsqueeze(torch.cat((torch.ones(pos_inter.shape[0]), torch.zeros(neg_inter.shape[0]))), 1)  # (2*batch,1)

        # inter
        inter = torch.cat((pos_inter, neg_inter), 0)  # (2*batch,1)
        lbl = lbl.to(inter.device)
        inter_loss = self.BCE_loss(inter, lbl)

        # intra_multi
        intra_multi = torch.cat((pos_multi, neg_multi), 0)  # (2*batch,1)
        intra_multi_loss = self.BCE_loss(intra_multi, lbl)

        # intra_pair
        pos_ = torch.sum(pos_matrix_, -1)  # (batch, all_len)
        neg_ = torch.sum(neg_matrix_, -1)  # (batch, all_len)
        intra_pair = - torch.log((pos_ + 1e-10) / (pos_ + neg_ + 1e-10))  # (batch,all_len)
        intra_pair = intra_pair * pos_entity_mask # (batch,all_len)
        intra_pair_loss = torch.sum(intra_pair, -1)  # (batch)
        intra_pair_loss = intra_pair_loss / torch.sum(pos_entity_mask, -1)

        inter_loss = torch.mean(inter_loss)
        intra_multi_loss = torch.mean(intra_multi_loss)
        intra_pair_loss = torch.mean(intra_pair_loss)

        loss = self.weight_inter * inter_loss + self.weight_intra_multi * intra_multi_loss + self.weight_intra_pair * intra_pair_loss

        if self.training:
            return loss
        else:
            inter_score = self.sigmoid(pos_inter)  # (batch,1)
            inter_score = torch.squeeze(inter_score) # (batch)

            intra_multi_score = self.sigmoid(pos_multi)  # (batch,1)
            intra_multi_score = torch.squeeze(intra_multi_score) # (batch)

            mask = torch.eye(pos_matrix.shape[1]) * 10000000.0  # (all_len,all_len)
            mask = torch.unsqueeze(mask, 0) # (1,all_len,all_len)
            mask = mask.to(pos_matrix.device)
            mask_ = torch.unsqueeze(pos_entity_mask, -1) # (batch,all_len,1)
            mask_ = torch.matmul(mask_, torch.transpose(mask_, 1, 2)) # (batch,all_len,all_len)
            mask_ = - mask_ + 1.0 # (batch,all_len,all_len)
            mask_ = mask_ * 10000000.0

            pos_matrix_ = pos_matrix + mask_ + mask
            intra_pair_score = []
            for i in range(pos_matrix_.shape[0]):
                _ = torch.min(pos_matrix_[i])
                intra_pair_score.append(_)

            intra_pair_score = torch.FloatTensor(intra_pair_score).to(pos_matrix.device) #
            intra_pair_score = self.sigmoid(intra_pair_score)

            score = - self.weight_inter * inter_score - self.weight_intra_multi * intra_multi_score - self.weight_intra_pair * intra_pair_score # (batch)

            return score



class Attention(nn.Module):
    def __init__(self, center_type, context_type, out_dim):
        super().__init__()
        self.center_type = center_type
        self.bilinear = nn.ModuleDict()
        for tp in context_type:
            self.bilinear[tp] = nn.Linear(out_dim, out_dim, bias=False)
            nn.init.xavier_normal_(self.bilinear[tp].weight, gain=1.414)

        self.leakyrelu = nn.LeakyReLU()
        self.softmax = nn.Softmax(dim=1)

        self.att_drop = nn.Dropout(0.5)

    def forward(self, event):
        center = event[self.center_type]  # (batch,1,out_dim)
        context_z = None
        context_k = None
        for key in event.keys():
            if key != self.center_type:
                context_k_ = self.bilinear[key](event[key])  # (batch,len[key],out_dim)
                if context_k is None:
                    context_z = event[key]
                    context_k = context_k_
                else:
                    context_z = torch.cat((context_z, event[key]), 1)  # (batch,all_len-1,out_dim)
                    context_k = torch.cat((context_k, context_k_), 1)  # (batch,all_len-1,out_dim)

        context_coef = context_k.matmul(torch.transpose(center, 1, 2))  # (batch,all_len-1,1)
        context_coef = torch.squeeze(context_coef)  # (batch,all_len-1)
        att = self.leakyrelu(context_coef)
        att = self.softmax(att)  # (batch,all_len-1)
        att = torch.unsqueeze(att, 2)  # (batch,all_len-1,1)
        context_aggre = (att * context_z).sum(dim=1)  # (batch,out_dim)
        cur_event = torch.cat((context_aggre, torch.squeeze(center)), 1)  # (batch,out_dim*2)

        return cur_event


class SelfAttention(nn.Module):
    def __init__(self, center_type, context_type, out_dim, num_of_attention_heads, type_n):
        super().__init__()

        self.num_attention_heads = num_of_attention_heads
        self.attention_head_size = int(out_dim / num_of_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(out_dim, self.all_head_size)
        self.key = nn.Linear(out_dim, self.all_head_size)
        self.value = nn.Linear(out_dim, self.all_head_size)

        self.dense = nn.Linear(self.all_head_size, out_dim)

        self.type_embedding = nn.Embedding(type_n, out_dim, padding_idx=0)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, hidden_states, type_num):  # (batch,all_len-1,out_dim)
        type_embedding = self.type_embedding(type_num) # (batch,all_len-1,out_dim)
        hidden_states = hidden_states + type_embedding # (batch,all_len-1,out_dim)

        mixed_query_layer = self.query(hidden_states)  # (batch,all_len-1,all_head_size)
        mixed_key_layer = self.key(hidden_states)  # (batch,all_len-1,all_head_size)
        mixed_value_layer = self.value(hidden_states)  # (batch,all_len-1,all_head_size)

        query_layer = self.transpose_for_scores(mixed_query_layer)  # (batch,num_of_head,all_len-1,head_size)
        key_layer = self.transpose_for_scores(mixed_key_layer)  # (batch,num_of_head,all_len-1,head_size)
        value_layer = self.transpose_for_scores(mixed_value_layer)  # (batch,num_of_head,all_len-1,head_size)

        attention_scores = torch.matmul(query_layer,
                                        key_layer.transpose(-1, -2))  # (batch,num_of_head,all_len-1,all_len-1)
        attention_scores = attention_scores / math.sqrt(
            self.attention_head_size)  # (batch,num_of_head,all_len-1,all_len-1)
        attention_probs = nn.Softmax(dim=-1)(attention_scores)  # (batch,num_of_head,all_len-1,all_len-1)
        context_layer = torch.matmul(attention_probs, value_layer)  # (batch,num_of_head,all_len-1,head_size)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()  # (batch,all_len-1,num_of_head,head_size)

        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)  # (batch,all_len-1,all_head_size)

        output = self.dense(context_layer)  # (batch,all_len-1,out_dim)

        return output



import os
from dgl.data.utils import download, extract_archive
from dgl.data import DGLDataset
from dgl.data.utils import load_graphs
from dgl.data.utils import save_graphs


# Download and process the dataset for mg2vec

class Mg2vecDataSet(DGLDataset):
    _prefix = 'https://s3.cn-north-1.amazonaws.com.cn/dgl-data/'
    _urls = {
        'dblp4Mg2vec_4': 'dataset/openhgnn/dblp4Mg2vec_4.zip',
        'dblp4Mg2vec_5': 'dataset/openhgnn/dblp4Mg2vec_5.zip',
    }

    def __init__(self, name, raw_dir=None, force_reload=False, verbose=False):
        assert name in ['dblp4Mg2vec_4', 'dblp4Mg2vec_5']
        raw_dir = './openhgnn/dataset'
        self.data_path = './openhgnn/dataset/' + name + '.zip'
        self.g_path = './openhgnn/dataset/' + name + '/graph.bin'
        url = self._prefix + self._urls[name]
        super(Mg2vecDataSet, self).__init__(name=name,
                                            url=url,
                                            raw_dir=raw_dir,
                                            force_reload=force_reload,
                                            verbose=verbose)

    def download(self):
        if os.path.exists(self.data_path):
            pass
        else:
            file_path = os.path.join(self.raw_dir)
            download(self.url, path=file_path)
        tem_path = os.path.join(self.raw_dir, self.name)
        extract_archive(self.data_path, tem_path)

    def process(self):
        g, _ = load_graphs(self.g_path)
        self._g = g[0]

    def __getitem__(self, idx):
        # get one example by index
        assert idx == 0, "This dataset has only one graph"
        return self._g

    def __len__(self):
        # number of data examples
        return 1

    def save(self):
        # save processed data to directory `self.save_path`
        pass

    def load(self):
        # load processed data from directory `self.save_path`
        pass

    def has_cache(self):
        # check whether there are processed data in `self.save_path`
        pass



# TODO: This file should be deprecated, it's just for testing.
# TODO: TEST!!!
import sys
import os
sys.path.append('/home/ubuntu/openhgnn/')
os.chdir('/home/ubuntu/openhgnn/')

from openhgnn.models.MAGNN import MAGNN
from openhgnn.sampler.MAGNN_sampler import MAGNN_sampler, collate_fn
from openhgnn.sampler.test_config import CONFIG
import argparse
import warnings
import time
import dgl
import numpy as np
import torch as th
import torch.nn.functional as F
from torch.utils.data import DataLoader
from sklearn.metrics import f1_score
from sklearn.model_selection import train_test_split
from sklearn.svm import LinearSVC

class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""
    def __init__(self, patience, verbose=False, delta=0, save_path='checkpoint.pt'):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        self.save_path = save_path

    def __call__(self, val_loss, model):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
        elif score < self.best_score - self.delta:
            self.counter += 1
            print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            self.counter = 0

    def save_checkpoint(self, val_loss, model):
        """Saves model when validation loss decrease."""
        if self.verbose:
            print(f'Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ...')
        th.save(model.state_dict(), self.save_path)
        self.val_loss_min = val_loss

def load_hg(args):
    hg_dir = 'openhgnn/dataset/'
    hg,_ = dgl.load_graphs(hg_dir+'{}/graph.bin'.format(args.dataset), [0])
    hg = hg[0]
    return hg

def svm_test(X, y, test_sizes=(0.2, 0.4, 0.6, 0.8), repeat=10):
    # This method is implemented by author
    random_states = [182318 + i for i in range(repeat)]
    result_macro_f1_list = []
    result_micro_f1_list = []
    for test_size in test_sizes:
        macro_f1_list = []
        micro_f1_list = []
        for i in range(repeat):
            X_train, X_test, y_train, y_test = train_test_split(
                X, y, test_size=test_size, shuffle=True, random_state=random_states[i])
            svm = LinearSVC(dual=False)
            svm.fit(X_train, y_train)
            y_pred = svm.predict(X_test)
            macro_f1 = f1_score(y_test, y_pred, average='macro')
            micro_f1 = f1_score(y_test, y_pred, average='micro')
            macro_f1_list.append(macro_f1)
            micro_f1_list.append(micro_f1)
        result_macro_f1_list.append([np.mean(macro_f1_list), np.std(macro_f1_list)])
        result_micro_f1_list.append([np.mean(micro_f1_list), np.std(micro_f1_list)])
    return result_macro_f1_list, result_micro_f1_list

if __name__ == '__main__':

    warnings.filterwarnings('ignore')
    args = argparse.Namespace(**CONFIG)
    hg = load_hg(args)
    train_mask = hg.nodes[args.category].data['train_mask']
    val_mask = hg.nodes[args.category].data['val_mask']
    test_mask = hg.nodes[args.category].data['test_mask']
    model = MAGNN.build_model_from_args(args, hg)
    print(args)

    sampler = MAGNN_sampler(g=hg, mask=train_mask.cpu().numpy(), num_layers=args.num_layers, category=args.category,
                            metapath_list=model.metapath_list, num_samples=args.num_samples, dataset_name=args.dataset)

    dataloader = DataLoader(dataset=sampler, batch_size=args.batch_size, shuffle=True, num_workers=args.num_workers,
                            collate_fn=collate_fn, drop_last=False)

    optimizer = th.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.weight_decay)
    early_stop = EarlyStopping(patience=args.patience, verbose=True,
                               save_path='/home/ubuntu/openhgnn/openhgnn/sampler/checkpoint/checkpoint_dblp.pt')
    model = model.to(args.device)

    for epoch in range(args.max_epoch):
        sampler.mask = train_mask
        t = time.perf_counter()
        model.train()
        print("...Start the mini batch training...")
        for num_iter, (sub_g, mini_mp_inst, seed_nodes) in enumerate(dataloader):
            print("Sampling {} seed_nodes with duration(s): {}".format(len(seed_nodes[args.category]), time.perf_counter() - t))
            model.mini_reset_params(mini_mp_inst)
            sub_g = sub_g.to(args.device)
            pred, _ = model(sub_g)
            pred = pred[args.category][seed_nodes[args.category]]
            lbl = sub_g.nodes[args.category].data['labels'][seed_nodes[args.category]]
            loss = F.cross_entropy(pred, lbl)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            print("Iter:{}, the training loss of this batch is {}".format(
                num_iter, loss.item()
            ))
            t = time.perf_counter()

        print()
        model.eval()
        with th.no_grad():
            sampler.mask = val_mask
            val_loader = DataLoader(dataset=sampler, batch_size=args.batch_size, shuffle=False,
                                    num_workers=args.num_workers, collate_fn=collate_fn, drop_last=False)

            logp_val_all = []
            embed_val_all = []
            for num_iter, (sub_g, mini_mp_inst, seed_nodes) in enumerate(val_loader):
                sub_g = sub_g.to(args.device)
                model.mini_reset_params(mini_mp_inst)
                pred_val, embed_val = model(sub_g)
                pred_val = pred_val[args.category][seed_nodes[args.category]]
                embed_val = embed_val[args.category][seed_nodes[args.category]]
                logp_val = F.log_softmax(pred_val, 1)
                logp_val_all.append(logp_val)
                embed_val_all.append(embed_val.cpu().numpy())

            lbl_val = hg.nodes[args.category].data['labels'][val_mask]
            lbl_val = lbl_val.cuda()
            embed_val_all = np.concatenate(embed_val_all, 0)
            loss_val = F.nll_loss(th.cat(logp_val_all, 0), lbl_val)
            lbl_val = lbl_val.cpu().numpy()
            val_f1_macro_list, val_f1_micro_list = svm_test(embed_val_all, lbl_val)

            print('Epoch {}. val_loss is {}'.format(epoch, loss_val))
            print('Macro-F1: ' + ', '.join(['{:.6f}~{:.6f} ({:.1f})'.format(
                macro_f1[0], macro_f1[1], train_size) for macro_f1, train_size in
                zip(val_f1_macro_list, [0.8, 0.6, 0.4, 0.2])]))
            print('Micro-F1: ' + ', '.join(['{:.6f}~{:.6f} ({:.1f})'.format(
                micro_f1[0], micro_f1[1], train_size) for micro_f1, train_size in
                zip(val_f1_micro_list, [0.8, 0.6, 0.4, 0.2])]))
            early_stop(loss_val, model)
            if early_stop.early_stop:
                print("Early Stopping!")
                break

    print('----------TEST-----------')
    model.eval()  # Test on full graphprint()
    with th.no_grad():
        sampler.mask = test_mask
        test_loader = DataLoader(dataset=sampler, batch_size=args.batch_size, shuffle=False,
                                num_workers=args.num_workers, collate_fn=collate_fn, drop_last=False)

        logp_test_all = []
        embed_test_all = []
        for num_iter, (sub_g, mini_mp_inst, seed_nodes) in enumerate(test_loader):
            sub_g = sub_g.to(args.device)
            model.mini_reset_params(mini_mp_inst)
            pred_test, embed_test = model(sub_g)
            pred_test = pred_test[args.category][seed_nodes[args.category]]
            embed_test = embed_test[args.category][seed_nodes[args.category]]
            logp_test = F.log_softmax(pred_test, 1)
            logp_test_all.append(logp_test)
            embed_test_all.append(embed_test.cpu().numpy())

        lbl_test = hg.nodes[args.category].data['labels'][test_mask]
        lbl_test = lbl_test.cuda()
        embed_test_all = np.concatenate(embed_test_all, 0)
        loss_test = F.nll_loss(th.cat(logp_test_all, 0), lbl_test)
        lbl_test = lbl_test.cpu().numpy()
        test_f1_macro_list, test_f1_micro_list = svm_test(embed_test_all, lbl_test)

        print('Macro-F1: ' + ', '.join(['{:.6f}~{:.6f} ({:.1f})'.format(
            macro_f1[0], macro_f1[1], train_size) for macro_f1, train_size in
            zip(test_f1_macro_list, [0.8, 0.6, 0.4, 0.2])]))
        print('Micro-F1: ' + ', '.join(['{:.6f}~{:.6f} ({:.1f})'.format(
            micro_f1[0], micro_f1[1], train_size) for micro_f1, train_size in
            zip(test_f1_micro_list, [0.8, 0.6, 0.4, 0.2])]))



import torch as th
from . import BaseDataset, register_dataset
from . import Mg2vecDataSet


@register_dataset('edge_classification')
class EdgeClassificationDataset(BaseDataset):
    r"""
        The class *EdgeClassificationDataset* is a base class for datasets which can be used in task *edge classification*.
        So its subclass should contain attributes such as graph, category, num_classes and so on.
        Besides, it should implement the functions *get_labels()* and *get_split()*.

        Attributes
        -------------
        g : dgl.DGLHeteroGraph
            The heterogeneous graph.
        category : str
            The category(or target) node type need to be predict. In general, we predict only one node type.
        num_classes : int
            The target node  will be classified into num_classes categories.
        has_feature : bool
            Whether the dataset has feature. Default ``False``.
        multi_label : bool
            Whether the node has multi label. Default ``False``. For now, only HGBn-IMDB has multi-label.
        """
    def __init__(self, *args, **kwargs):
        super(EdgeClassificationDataset, self).__init__(*args, **kwargs)
        self.g = None
        self.category = None
        self.num_classes = None
        self.has_feature = False
        self.multi_label = False

    def get_labels(self):
        r"""
                The subclass of dataset should overwrite the function. We can get labels of target nodes through it.

                Notes
                ------
                In general, the labels are th.LongTensor.
                But for multi-label dataset, they should be th.FloatTensor. Or it will raise
                RuntimeError: Expected object of scalar type Long but got scalar type Float for argument #2 target' in call to _thnn_nll_loss_forward

                return
                -------
                labels : torch.Tensor
                """
        if 'labels' in self.g.edges[self.category].data:
            labels = self.g.edges[self.category].data.pop('labels').long()
        elif 'label' in self.g.edges[self.category].data:
            labels = self.g.edges[self.category].data.pop('label').long()
        else:
            raise ValueError('Labels of nodes are not in the hg.edges[category].data.')
        labels = labels.float() if self.multi_label else labels
        return labels

    def get_split(self, validation=True):
        r"""

                Parameters
                ----------
                validation : bool
                    Whether to split dataset. Default ``True``. If it is False, val_idx will be same with train_idx.

                We can get idx of train, validation and test through it.

                return
                -------
                train_idx, val_idx, test_idx : torch.Tensor, torch.Tensor, torch.Tensor
                """
        if 'train_mask' not in self.g.edges[self.category].data:
            self.logger.dataset_info("The dataset has no train mask. "
                                     "So split the category nodes randomly. And the ratio of train/test is 8:2.")
            num_nodes = self.g.number_of_nodes(self.category)
            n_test = int(num_nodes * 0.2)
            n_train = num_nodes - n_test

            train, test = th.utils.data.random_split(range(num_nodes), [n_train, n_test])
            train_idx = th.tensor(train.indices)
            test_idx = th.tensor(test.indices)
            if validation:
                self.logger.dataset_info("Split train into train/valid with the ratio of 8:2 ")
                random_int = th.randperm(len(train_idx))
                valid_idx = train_idx[random_int[:len(train_idx) // 5]]
                train_idx = train_idx[random_int[len(train_idx) // 5:]]
            else:
                self.logger.dataset_info("Set valid set with train set.")
                valid_idx = train_idx
                train_idx = train_idx
        else:
            train_mask = self.g.edges[self.category].data.pop('train_mask')
            test_mask = self.g.edges[self.category].data.pop('test_mask')
            train_idx = th.nonzero(train_mask, as_tuple=False).squeeze()
            test_idx = th.nonzero(test_mask, as_tuple=False).squeeze()
            if validation:
                if 'val_mask' in self.g.edges[self.category].data:
                    val_mask = self.g.edges[self.category].data.pop('val_mask')
                    valid_idx = th.nonzero(val_mask, as_tuple=False).squeeze()
                elif 'valid_mask' in self.g.edges[self.category].data:
                    val_mask = self.g.edges[self.category].data.pop('valid_mask').squeeze()
                    valid_idx = th.nonzero(val_mask, as_tuple=False).squeeze()
                else:
                    self.logger.dataset_info("Split train into train/valid with the ratio of 8:2 ")
                    random_int = th.randperm(len(train_idx))
                    valid_idx = train_idx[random_int[:len(train_idx) // 5]]
                    train_idx = train_idx[random_int[len(train_idx) // 5:]]
            else:
                self.logger.dataset_info("Set valid set with train set.")
                valid_idx = train_idx
                train_idx = train_idx
        self.train_idx = train_idx
        self.valid_idx = valid_idx
        self.test_idx = test_idx
        return self.train_idx, self.valid_idx, self.test_idx


@register_dataset('hin_edge_classification')
class HIN_EdgeClassification(EdgeClassificationDataset):
    r"""
        The HIN dataset are all used in different papers. So we preprocess them and store them as form of dgl.DGLHeteroGraph.
        The dataset name combined with paper name through 4(for).

        Dataset Name :
        dblp4Mg2vec/ ...
        """
    def __init__(self, dataset_name, *args, **kwargs):
        super(HIN_EdgeClassification, self).__init__(*args, **kwargs)
        self.g, self.category, self.num_classes = self.load_HIN(dataset_name)

    def load_HIN(self, name_dataset):
        if name_dataset == 'dblp4Mg2vec_4':
            # which is used in MG2VEC with size=4
            dataset = Mg2vecDataSet(name='dblp4Mg2vec_4', raw_dir='')
            g = dataset[0].long()
            category = 'relation'
            num_classes = 3
            return g, category, num_classes
        if name_dataset == 'dblp4Mg2vec_5':
            # which is used in MG2VEC with size=5
            dataset = Mg2vecDataSet(name='dblp4Mg2vec_5', raw_dir='')
            g = dataset[0].long()
            category = 'relation'
            num_classes = 3
            return g, category, num_classes



import torch as t
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data as dataloader
import torch.optim as optim
import numpy as np
import datetime
import torch.utils.data as data
from openhgnn.trainerflow.base_flow import BaseFlow
from openhgnn.trainerflow import register_flow
from openhgnn.models import build_model
from ..tasks import build_task
from openhgnn.models.HGCL import HGCL

saveDefault = False

@register_flow('hgcltrainer')
class HGCLtrainer(BaseFlow):
    def __init__(self, args):
        super(HGCLtrainer, self).__init__(args)
        self.args = args

        self.task = build_task(args)
        self.hg = self.task.dataset.g
        self.userNum = self.hg.number_of_nodes('user')
        self.itemNum = self.hg.number_of_nodes('item')

        self.model = build_model(self.model).build_model_from_args(args=self.args, hg=self.hg).to(self.device)
        self.opt = optim.Adam(self.model.parameters(), lr=self.args.lr)

        trainMat = self.hg.adj_external(etype=('user', 'interact_train', 'item'), scipy_fmt='coo')
        testMat = self.hg.adj_external(etype=('user', 'interact_test', 'item'), scipy_fmt='coo')
        train_u, train_v, train_r = trainMat.row, trainMat.col, trainMat.data
        assert np.sum(train_r == 0) == 0
        test_u, test_v = testMat.row, testMat.col
        train_data = np.hstack((train_u.reshape(-1, 1), train_v.reshape(-1, 1))).tolist()
        test_data = np.hstack((test_u.reshape(-1, 1), test_v.reshape(-1, 1))).tolist()
        train_dataset = BPRData(train_data, self.itemNum, trainMat, 1, True)
        test_dataset = BPRData(test_data, self.itemNum, trainMat, 0, False)
        self.train_loader = dataloader.DataLoader(train_dataset, batch_size=self.args.batch, shuffle=True,
                                                  num_workers=0)
        self.test_loader = dataloader.DataLoader(test_dataset, batch_size=1024 * 1000, shuffle=False, num_workers=0)
        self.train_losses = []
        self.test_hr = []
        self.test_ndcg = []

    def predictModel(self, user, pos_i, neg_j, isTest=False):
        if isTest:
            pred_pos = t.sum(user * pos_i, dim=1)
            return pred_pos
        else:
            pred_pos = t.sum(user * pos_i, dim=1)
            pred_neg = t.sum(user * neg_j, dim=1)
            return pred_pos, pred_neg

    # Contrastive Learning
    def ssl_loss(self, data1, data2, index):
        index = t.unique(index)
        embeddings1 = data1[index]
        embeddings2 = data2[index]
        norm_embeddings1 = F.normalize(embeddings1, p=2, dim=1)
        norm_embeddings2 = F.normalize(embeddings2, p=2, dim=1)
        pos_score = t.sum(t.mul(norm_embeddings1, norm_embeddings2), dim=1)
        all_score = t.mm(norm_embeddings1, norm_embeddings2.T)
        pos_score = t.exp(pos_score / self.args.ssl_temp)
        all_score = t.sum(t.exp(all_score / self.args.ssl_temp), dim=1)
        ssl_loss = (-t.sum(t.log(pos_score / ((all_score)))) / (len(index)))
        return ssl_loss

    # Model train
    def _mini_train_step(self):
        epoch_loss = 0
        self.train_loader.dataset.ng_sample()
        step_num = 0  # count batch num
        for user, item_i, item_j in self.train_loader:
            user = user.long().cuda()
            item_i = item_i.long().cuda()
            item_j = item_j.long().cuda()
            step_num += 1
            self.istrain = True
            itemindex = t.unique(t.cat((item_i, item_j)))
            userindex = t.unique(user)
            self.userEmbed, self.itemEmbed, self.ui_userEmbedall, self.ui_itemEmbedall, self.ui_userEmbed, self.ui_itemEmbed, metaregloss = self.model(
                self.istrain, userindex, itemindex, norm=1)

            # Contrastive Learning of collaborative relations
            ssl_loss_user = self.ssl_loss(self.ui_userEmbed, self.userEmbed, user)
            ssl_loss_item = self.ssl_loss(self.ui_itemEmbed, self.itemEmbed, item_i)
            ssl_loss = self.args.ssl_ureg * ssl_loss_user + self.args.ssl_ireg * ssl_loss_item

            # prediction
            pred_pos, pred_neg = self.predictModel(self.ui_userEmbedall[user], self.ui_itemEmbedall[item_i],
                                                   self.ui_itemEmbedall[item_j])
            bpr_loss = - nn.LogSigmoid()(pred_pos - pred_neg).sum()
            epoch_loss += bpr_loss.item()
            regLoss = (t.norm(self.ui_userEmbedall[user]) ** 2 + t.norm(self.ui_itemEmbedall[item_i]) ** 2 + t.norm(
                self.ui_itemEmbedall[item_j]) ** 2)
            loss = ((bpr_loss + regLoss * self.args.reg) / self.args.batch) + ssl_loss * self.args.ssl_beta + metaregloss * self.args.metareg

            self.opt.zero_grad()
            loss.backward()
            nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=20, norm_type=2)
            self.opt.step()
        return epoch_loss

    def test(self):
        HR = []
        NDCG = []

        with t.no_grad():
            uid = np.arange(0, self.userNum)
            iid = np.arange(0, self.itemNum)
            self.istrain = False
            _, _, self.ui_userEmbed, self.ui_itemEmbed, _, _, _ = self.model(self.istrain, uid, iid, norm=1)
            for test_u, test_i in self.test_loader:
                test_u = test_u.long().cuda()
                test_i = test_i.long().cuda()
                pred = self.predictModel(self.ui_userEmbed[test_u], self.ui_itemEmbed[test_i], None, isTest=True)
                batch = int(test_u.cpu().numpy().size / 100)
                for i in range(batch):
                    batch_socres = pred[i * 100:(i + 1) * 100].view(-1)
                    _, indices = t.topk(batch_socres, self.args.topk)
                    tmp_item_i = test_i[i * 100:(i + 1) * 100]
                    recommends = t.take(tmp_item_i, indices).cpu().numpy().tolist()
                    gt_item = tmp_item_i[0].item()
                    HR.append(self.hit(gt_item, recommends))
                    NDCG.append(self.ndcg(gt_item, recommends))
        return np.mean(HR), np.mean(NDCG)


    def hit(self, gt_item, pred_items):
        if gt_item in pred_items:
            return 1
        return 0

    def ndcg(self, gt_item, pred_items):
        if gt_item in pred_items:
            index = pred_items.index(gt_item)
            return np.reciprocal(np.log2(index + 2))
        return 0

    def log(self, msg, save=None, oneline=False):
        global logmsg
        global saveDefault
        time = datetime.datetime.now()
        tem = '%s: %s' % (time, msg)
        if save != None:
            if save:
                logmsg += tem + '\n'
        elif saveDefault:
            logmsg += tem + '\n'
        if oneline:
            print(tem, end='\r')
        else:
            print(tem)

    def _full_train_setp(self):
        pass

    def _test_step(self, split=None, logits=None):
        pass

    def train(self):
        # self.prepareModel()
        self.curEpoch = 0
        best_hr = -1
        best_ndcg = -1
        best_epoch = -1
        HR_lis = []
        for e in range(self.args.epochs + 1):
            self.curEpoch = e
            # train
            self.log("**************************************************************")
            epoch_loss = self._mini_train_step()
            self.train_losses.append(epoch_loss)
            self.log("epoch %d/%d, epoch_loss=%.2f" % (e, self.args.epochs, epoch_loss))

            # test
            HR, NDCG = self.test()  #
            self.test_hr.append(HR)
            self.test_ndcg.append(NDCG)
            self.log("epoch %d/%d, HR@10=%.4f, NDCG@10=%.4f" % (e, self.args.epochs, HR, NDCG))
            # self.adjust_learning_rate()
            if HR > best_hr:
                best_hr, best_ndcg, best_epoch = HR, NDCG, e

            HR_lis.append(HR)

        print("*****************************")
        self.log("best epoch = %d, HR= %.4f, NDCG=%.4f" % (best_epoch, best_hr, best_ndcg))
        print("*****************************")
        print(self.args)

class BPRData(data.Dataset):
    def __init__(self, data,
                 num_item, train_mat=None, num_ng=0, is_training=None):
        super(BPRData, self).__init__()
        """ Note that the labels are only useful when training, we thus 
            add them in the ng_sample() function.
        """
        self.data = np.array(data)
        self.num_item = num_item
        self.train_mat = train_mat
        self.num_ng = num_ng
        self.is_training = is_training

    def ng_sample(self):
        assert self.is_training, 'no need to sampling when testing'
        tmp_trainMat = self.train_mat.todok()
        length = self.data.shape[0]
        self.neg_data = np.random.randint(low=0, high=self.num_item, size=length)

        for i in range(length):
            uid = self.data[i][0]
            iid = self.neg_data[i]
            if (uid, iid) in tmp_trainMat:
                while (uid, iid) in tmp_trainMat:
                    iid = np.random.randint(low=0, high=self.num_item)
                self.neg_data[i] = iid

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        user = self.data[idx][0]
        item_i = self.data[idx][1]
        if self.is_training:
            neg_data = self.neg_data
            item_j = neg_data[idx]
            return user, item_i, item_j
        else:
            return user, item_i



import copy
import torch.nn.functional as F
import torch.nn as nn
import torch
from openhgnn.models import BaseModel,register_model
# from torch_geometric.nn import GCNConv
from dgl.nn.pytorch import GraphConv

class EMA():
    def __init__(self, beta):
        super().__init__()
        self.beta = beta

    def update_average(self, old, new):
        if old is None:
            return new
        return old * self.beta + (1 - self.beta) * new

class MLP(nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, num_layers,
                 dropout=.0):
        super(MLP, self).__init__()
        self.linear1 = nn.Linear(in_channels, hidden_channels)
        self.linear2 = nn.Linear(hidden_channels, out_channels)
        self.dropout = dropout

    def reset_parameters(self):
        self.linear1.reset_parameters()
        self.linear2.reset_parameters()

    def forward(self, data):
        x = data
        x = F.relu(self.linear1(x))
        x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.linear2(x)
        return x

def set_requires_grad(model, val):
    for p in model.parameters():
        p.requires_grad = val

def loss_fn(x, y):
    x = F.normalize(x, dim=-1, p=2)
    y = F.normalize(y, dim=-1, p=2)
    return 2 - 2 * (x * y).sum(dim=-1)

def update_moving_average(target_ema_updater, ma_model, current_model):
    for current_params, ma_params in zip(current_model.parameters(), ma_model.parameters()):
        old_weight, up_weight = ma_params.data, current_params.data
        ma_params.data = target_ema_updater.update_average(old_weight, up_weight)

class GCN(nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, num_layers=2,
                 dropout=0.5, device='cpu', save_mem=False, use_bn=True):
        super(GCN, self).__init__()
        cached = False
        add_self_loops = True
        self.convs = nn.ModuleList()
        self.convs.append(
            # GCNConv(in_channels, hidden_channels, cached=cached, normalize=not save_mem, add_self_loops=add_self_loops)
            GraphConv(in_channels, hidden_channels, allow_zero_in_degree=False)
        )

        self.bns = nn.ModuleList()
        self.bns.append(nn.BatchNorm1d(hidden_channels))
        for _ in range(num_layers - 2):
            self.convs.append(
                # GCNConv(hidden_channels, hidden_channels, cached=cached, normalize=not save_mem,
                #         add_self_loops=add_self_loops)
                GraphConv(hidden_channels, hidden_channels, allow_zero_in_degree=False)
            )

            self.bns.append(nn.BatchNorm1d(hidden_channels))

        self.convs.append(
            # GCNConv(hidden_channels, out_channels, cached=cached, normalize=not save_mem,
            #         add_self_loops=add_self_loops)
            GraphConv(hidden_channels, out_channels, allow_zero_in_degree=False)
        )

        self.dropout = dropout
        self.activation = F.relu
        self.use_bn = use_bn
        self.device = device

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()
        for bn in self.bns:
            bn.reset_parameters()

    def forward(self, data):
        x = data.g.ndata['feat']
        edges = data.g.edges()
        edges_tensor = torch.cat([edges[0].unsqueeze(0), edges[1].unsqueeze(0)], 0)
        x = x.to(self.device)
        edges_tensor = edges_tensor.to(self.device)
        graph = data.g.to(self.device)
        for i, conv in enumerate(self.convs[:-1]):
            # x = conv(x, edges_tensor)
            x = conv(graph,x)
            if self.use_bn:
                x = self.bns[i](x)
            x = self.activation(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        # x = self.convs[-1](x, edges_tensor)
        x = self.convs[-1](graph,x)
        return x

class LogisticRegression(nn.Module):
    def __init__(self, num_dim, num_class):
        super().__init__()
        self.linear = nn.Linear(num_dim, num_class)
        torch.nn.init.xavier_uniform_(self.linear.weight.data)
        self.linear.bias.data.fill_(0.0)
        self.cross_entropy = nn.CrossEntropyLoss()

    def forward(self, x, y):
        logits = self.linear(x)
        loss = self.cross_entropy(logits, y)
        return logits, loss

@register_model('DSSL')
class DSSL(BaseModel):

    @classmethod
    def build_model_from_args(cls, args, hg):
        if args.encoder == 'GCN':
            encoder = GCN(in_channels=args.feature_dim,
                          hidden_channels=args.hidden_channels,
                          out_channels=args.hidden_channels,
                          num_layers=args.num_layers, use_bn=not args.no_bn,
                          dropout=args.dropout,
                          device = args.device).to(args.device)
        return cls(encoder,args.hidden_channels,args.dataset,args.device,args.cluster_num,args.alpha,args.gamma,args.tao,
            args.beta,args.tau)

    def __init__(self,encoder, hidden_channels,dataset, device, cluster_num, alpha,gamma,tao,beta,moving_average_decay=0.0):
        super(DSSL, self).__init__()
        self.dataset = dataset
        self.device = device
        self.cluster_num = cluster_num
        self.alpha = alpha
        self.gamma = gamma
        self.tao = tao
        self.beta = beta
        self.inner_dropout = True
        self.inner_activation = True
        self.online_encoder = encoder
        self.target_encoder = copy.deepcopy(self.online_encoder)
        set_requires_grad(self.target_encoder, False)
        self.target_ema_updater = EMA(moving_average_decay)

        self.mlp_inference = MLP(hidden_channels, hidden_channels, cluster_num, 2)
        self.mlp_predictor = MLP(hidden_channels, hidden_channels, hidden_channels, 2)
        self.clusters = nn.Parameter(nn.init.normal_(torch.Tensor(hidden_channels, cluster_num)))
        self.mlp_predictor2 = MLP(cluster_num, hidden_channels, hidden_channels, 1)

        self.Embedding_mlp = True
        self.inference_mlp = True

    def reset_parameters(self):
        self.mlp_inference.reset_parameters()
        self.mlp_predictor.reset_parameters()
        self.mlp_predictor2.reset_parameters()
        self.online_encoder.reset_parameters()
        self.target_encoder = copy.deepcopy(self.online_encoder)

    def forward(self, embedding,neighbor_embedding):
        # expand node embedding
        embedding_more = embedding[:, None, :]
        embedding_expand = embedding_more.expand(-1, neighbor_embedding.shape[1], -1)
        # loss
        if self.inference_mlp == True:  # self.inference_mlp = True
            k, k_node, entropy_loss = self.inference_network(embedding_expand, neighbor_embedding)
        # else:
        #     k, k_node, entropy_loss = self.inference_network2(embedding_expand, neighbor_embedding)

        if self.Embedding_mlp == True:  # self.Embedding_mlp = True
            main_loss = self.generative_network(embedding_expand, k, neighbor_embedding)
        # else:
        #     main_loss = self.generative_network2(embedding_expand, k, neighbor_embedding)

        context_loss = self.context_network(embedding, k_node)

        return main_loss, context_loss, entropy_loss, k_node

        # N × K, N × 10 * K, k is N*32
    def inference_network(self, embedding_expand, neighbor_embedding):
        # get k
        cat_embedding = embedding_expand * neighbor_embedding
        # cat_embedding=torch.cat((embedding_expand,neighbor_embedding))
        k = F.softmax(self.mlp_inference(cat_embedding), dim=2)
        k_node = k.mean(dim=1)  # to get P(k|x)
        negative_entropy = k_node * torch.log(k_node + 1e-10)
        # minimize negative entropy
        entropy_loss = negative_entropy.sum(-1).mean()
        return k, k_node, entropy_loss

    def generative_network(self, embedding_expand, k, neighbor_embedding):
        #
        # re-parameterization trick
        gumbel_k = F.gumbel_softmax(k, hard=False)
        central=self.mlp_predictor(embedding_expand)+ self.beta*self.mlp_predictor2(gumbel_k)
        neighbor=neighbor_embedding
        loss= loss_fn(central, neighbor.detach()).mean()

        return loss

    def context_network(self, embedding, k_node):
        # self.clusters.data = F.normalize(self.clusters.data, dim=-1, p=2)
        kprior = torch.matmul(embedding, self.clusters)         # 【1024，6】
        kprior = F.softmax(kprior/self.tao, dim=1)

        context_loss = k_node *torch.log(kprior+1e-10)          # 【1024，6】
        context_loss = - 1.0 * context_loss.sum(-1).mean()
        return context_loss

    def update_moving_average(self):
        #assert self.use_momentum, 'you do not need to update the moving average, since you have turned off momentum for the target encoder'
        assert self.target_encoder is not None, 'target encoder has not been created yet'
        update_moving_average(self.target_ema_updater,self.target_encoder, self.online_encoder)

    def update_cluster(self, new_center,batch_sum):
        with torch.no_grad():
            out_ids = torch.arange(self.cluster_num).to(self.device)
            out_ids = out_ids.long()
            self.clusters.index_copy_(1, out_ids, new_center)
            self.clusters.data=torch.mul(self.clusters.data.T ,1/ (batch_sum+1)).T


    def extra_loss(self):
        pass

    def get_emb(self):
        pass




import torch.nn as nn
import torch.nn.functional as F
import torch
import dgl
from dgl.nn.pytorch import GATConv
from . import BaseModel, register_model
from ..layers.macro_layer.SemanticConv import SemanticAttention
from ..layers.MetapathConv import MetapathConv
from ..utils.utils import extract_metapaths, get_ntypes_from_canonical_etypes



@register_model('SHGP')
class ATT_HGCN(BaseModel):

    @classmethod
    def build_model_from_args(cls, args, hg):
        ntypes = set()
        return cls(
                net_schema=args.net_schema,
                layer_shape=args.layer_shape,
                label_keys=list(args.label.keys()),
                type_fusion=args.type_fusion,
                type_att_size=args.type_att_size,
        )

    def __init__(self, net_schema, layer_shape, label_keys, type_fusion='att', type_att_size=64):
        super(ATT_HGCN, self).__init__()
        self.hgc1 = HeteGCNLayer(net_schema, layer_shape[0], layer_shape[1], type_fusion, type_att_size)
        self.hgc2 = HeteGCNLayer(net_schema, layer_shape[1], layer_shape[2], type_fusion, type_att_size)

        self.embd2class = nn.ParameterDict()
        self.bias = nn.ParameterDict()
        self.label_keys = label_keys
        self.layer_shape = layer_shape
        for k in label_keys:
            self.embd2class[k] = nn.Parameter(torch.FloatTensor(layer_shape[-2][k], layer_shape[-1][k]))
            nn.init.xavier_uniform_(self.embd2class[k].data, gain=1.414)
            self.bias[k] = nn.Parameter(torch.FloatTensor(1, layer_shape[-1][k]))
            nn.init.xavier_uniform_(self.bias[k].data, gain=1.414)

    def ini_embd2class(self):
        for k in self.label_keys:
            nn.init.xavier_uniform_(self.embd2class[k].data, gain=1.414)
            nn.init.xavier_uniform_(self.bias[k].data, gain=1.414)

    def forward(self, ft_dict, adj_dict):
        attention_list = []
        x_dict, attention_dict = self.hgc1(ft_dict, adj_dict)
        attention_list.append((attention_dict))
        x_dict = self.non_linear(x_dict)
        x_dict = self.dropout_ft(x_dict, 0.5)

        x_dict, attention_dict = self.hgc2(x_dict, adj_dict)
        attention_list.append((attention_dict))

        logits = {}
        embd = {}
        for k in self.label_keys:
            embd[k] = x_dict[k]
            logits[k] = torch.mm(x_dict[k], self.embd2class[k]) + self.bias[k]
        return logits, embd, attention_list

    def non_linear(self, x_dict):
        y_dict = {}
        for k in x_dict:
            y_dict[k] = F.elu(x_dict[k])
        return y_dict

    def dropout_ft(self, x_dict, dropout):
        y_dict = {}
        for k in x_dict:
            y_dict[k] = F.dropout(x_dict[k], dropout, training=self.training)
        return y_dict\

class _ATT_HGCN(nn.Module):

    def __init__(self):
        pass

    def forward(self):
        pass

class HeteGCNLayer(nn.Module):

    def __init__(self, net_schema, in_layer_shape, out_layer_shape, type_fusion, type_att_size):
        super(HeteGCNLayer, self).__init__()
        self.net_schema = net_schema
        self.in_layer_shape = in_layer_shape
        self.out_layer_shape = out_layer_shape

        self.hete_agg = nn.ModuleDict()
        for k in net_schema:
            self.hete_agg[k] = HeteAggregateLayer(k, net_schema[k], in_layer_shape, out_layer_shape[k], type_fusion,
                                                  type_att_size)

    def forward(self, x_dict, adj_dict):
        attention_dict = {}
        ret_x_dict = {}
        for k in self.hete_agg.keys():
            ret_x_dict[k], attention_dict[k] = self.hete_agg[k](x_dict, adj_dict[k])

        return ret_x_dict, attention_dict


class HeteAggregateLayer(nn.Module):

    def __init__(self, curr_k, nb_list, in_layer_shape, out_shape, type_fusion, type_att_size):
        super(HeteAggregateLayer, self).__init__()

        self.nb_list = nb_list
        self.curr_k = curr_k
        self.type_fusion = type_fusion
        self.W_rel = nn.ParameterDict()
        for k in nb_list:
            try:
                self.W_rel[k] = nn.Parameter(torch.FloatTensor(in_layer_shape[k], out_shape))
            except KeyError as ke:
                self.W_rel[k] = nn.Parameter(torch.FloatTensor(in_layer_shape[self.curr_k], out_shape))
            finally:
                nn.init.xavier_uniform_(self.W_rel[k].data, gain=1.414)

        self.w_self = nn.Parameter(torch.FloatTensor(in_layer_shape[curr_k], out_shape))
        nn.init.xavier_uniform_(self.w_self.data, gain=1.414)

        self.bias = nn.Parameter(torch.FloatTensor(1, out_shape))
        nn.init.xavier_uniform_(self.bias.data, gain=1.414)

        if type_fusion == 'att':
            self.w_query = nn.Parameter(torch.FloatTensor(out_shape, type_att_size))
            nn.init.xavier_uniform_(self.w_query.data, gain=1.414)
            self.w_keys = nn.Parameter(torch.FloatTensor(out_shape, type_att_size))
            nn.init.xavier_uniform_(self.w_keys.data, gain=1.414)
            self.w_att = nn.Parameter(torch.FloatTensor(2 * type_att_size, 1))
            nn.init.xavier_uniform_(self.w_att.data, gain=1.414)

    def forward(self, x_dict, adj_dict):
        attention_curr_k = 0
        self_ft = torch.mm(x_dict[self.curr_k], self.w_self)

        nb_ft_list = [self_ft]
        nb_name = [self.curr_k + '_self']
        for k in self.nb_list:
            try:
                nb_ft = torch.mm(x_dict[k], self.W_rel[k])
            except KeyError as ke:
                nb_ft = torch.mm(x_dict[self.curr_k], self.W_rel[k])
            finally:
                nb_ft = torch.spmm(adj_dict[k], nb_ft)
                nb_ft_list.append(nb_ft)
                nb_name.append(k)

        if self.type_fusion == 'mean':
            agg_nb_ft = torch.cat([nb_ft.unsqueeze(1) for nb_ft in nb_ft_list], 1).mean(1)
            attention = []
        elif self.type_fusion == 'att':
            att_query = torch.mm(self_ft, self.w_query).repeat(len(nb_ft_list), 1)
            att_keys = torch.mm(torch.cat(nb_ft_list, 0), self.w_keys)
            att_input = torch.cat([att_keys, att_query], 1)
            att_input = F.dropout(att_input, 0.5, training=self.training)
            e = F.elu(torch.matmul(att_input, self.w_att))
            attention = F.softmax(e.view(len(nb_ft_list), -1).transpose(0, 1), dim=1)  # 4025*3
            agg_nb_ft = torch.cat([nb_ft.unsqueeze(1) for nb_ft in nb_ft_list], 1).mul(attention.unsqueeze(-1)).sum(1)

        output = agg_nb_ft + self.bias

        return output, attention



import torch
import torch.nn as nn


class LogReg(nn.Module):
    def __init__(self, ft_in, nb_classes):
        super(LogReg, self).__init__()
        self.fc = nn.Linear(ft_in, nb_classes)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, seq):
        ret = self.fc(seq)
        return ret



import os

import torch
from torch.utils.data import DataLoader
import torch.optim as optim
from ..models import build_model
from . import BaseFlow, register_flow
from ..sampler import mg2vec_sampler
import numpy as np


@register_flow('mg2vec_trainer')
class Mg2vecTrainer(BaseFlow):
    def __init__(self, args):
        super(Mg2vecTrainer, self).__init__(args)
        self.mg2vec_sampler = None
        self.dataloader = None
        self.model = None
        self.embeddings_file_path = os.path.join(self.args.output_dir, self.args.dataset + '_mg2vec_embeddings.npy')
        self.embeddings_file_path2 = os.path.join(self.args.output_dir, self.args.dataset + '_mg2vec_embeddings.txt')
        self.load_trained_embeddings = False

    def preprocess(self):
        input_file = "./openhgnn/dataset/{}/meta.txt".format(self.args.dataset)
        block_size = self.args.batch_size * 100000
        self.mg2vec_sampler = mg2vec_sampler.Mg2vecSampler(input_file, block_size, self.args.alpha)
        self.dataloader = DataLoader(self.mg2vec_sampler, batch_size=self.args.batch_size, shuffle=True,
                                     num_workers=self.args.num_workers,
                                     )
        self.args.node_num = self.mg2vec_sampler.data.node_count
        self.args.mg_num = self.mg2vec_sampler.data.mg_count
        self.args.unigram = self.mg2vec_sampler.data.unigram
        self.model = build_model(self.model_name).build_model_from_args(self.args, self.hg).to(self.device)

    def train(self):
        emb = self.load_embeddings()
        emb_dict = dict()
        for nId, node in self.mg2vec_sampler.data.node_reverse_dict.items():
            emb_dict[int(node)] = emb[nId]
        # todo: only supports edge classification now
        metric = {
            'test': self.task.downstream_evaluate(logits=self.get_edge_embed(emb=emb_dict), evaluation_metric='acc_f1')}
        self.logger.train_info(self.logger.metric2str(metric))
        # metric = {'test': self.task.evaluate(emb_dict)}
        # self.logger.train_info(self.logger.metric2str(metric))

    def load_embeddings(self):
        if not self.load_trained_embeddings or not os.path.exists(self.embeddings_file_path):
            self.train_embeddings()
        emb = np.load(self.embeddings_file_path)
        return emb

    def train_embeddings(self):
        self.preprocess()
        epoch_index = 1
        optimizer = optim.Adam(list(self.model.parameters()), lr=self.args.lr)
        average_loss = 0.0
        step = 0
        print("train start")
        while True:
            for i, sampled_batch in enumerate(self.dataloader):
                if len(sampled_batch) > 0:
                    train_a = sampled_batch[0].to(self.device)
                    train_b = sampled_batch[1].to(self.device)
                    train_label = sampled_batch[2].to(self.device)
                    train_freq = sampled_batch[3].reshape(-1, 1).to(self.device)
                    train_weight = sampled_batch[4].reshape(-1, 1).to(self.device)

                    optimizer.zero_grad()
                    loss = self.model.forward(train_a, train_b, train_label, train_freq, train_weight, self.device)
                    loss.backward()
                    optimizer.step()

                    average_loss += loss.item()
                    step += 1
                    if step > 0 and step % 10000 == 0:
                        average_loss /= 10000
                        print('Average loss at step ', step, ': ', average_loss)
                        average_loss = 0.0
            if self.mg2vec_sampler.data.epoch_end:
                print("epoch %d end" % epoch_index)
                epoch_index += 1
                self.mg2vec_sampler.data.epoch_end = False
                if epoch_index > self.args.max_epoch:
                    break
            self.mg2vec_sampler.data.read_block()

        print("total step: ", step)
        self.model.save_embedding_np(self.embeddings_file_path)
        self.model.save_embedding(self.mg2vec_sampler.data.node_reverse_dict, self.embeddings_file_path2)

    def get_edge_embed(self, emb):
        edge_embed = []
        g = self.hg
        u, v = g.edges()
        core1_dict = g.nodes['core1'].data['id2node'].cpu()
        core2_dict = g.nodes['core2'].data['id2node'].cpu()
        for i in range(len(u)):
            edge_embed.append(np.hstack([emb[int(core1_dict[u[i]])], emb[int(core2_dict[v[i]])]]))
        x = np.array(edge_embed)
        return x



from openhgnn import Experiment
from unittest import TestCase


class TestExperiment(TestCase):
    experiment = Experiment(model='RGCN', task='node_classification', dataset='imdb4GTN', gpu=-1, use_best_config=True)
    pass



import torch
import torch.nn as nn
from tqdm import tqdm
from . import BaseFlow, register_flow
from ..models import build_model


@register_flow("abnorm_event_detection")
class AbnormEventDetection(BaseFlow):
    """AbnormEventDetection flow

    Supported Model: AEHCL
    Supported Dataset: aminer4AEHCL

    """

    def __init__(self, args=None):
        super(AbnormEventDetection, self).__init__(args)
        self.event_label = self.task.dataset.event_label
        args.center_type, args.context_type = self.task.dataset.center_type, self.task.dataset.context_type
        args.in_dim = self.task.dataset.in_dim
        all_len = self.task.dataset.type_max_num[args.center_type]
        for tp in args.context_type:
            all_len += self.task.dataset.type_max_num[tp]
        args.all_len = all_len
        self.model = build_model(self.model).build_model_from_args(self.args).to(self.device)
        self.ap = self.task.get_evaluator('AP')
        self.auc = self.task.get_evaluator('AUC')
        self.batch_size = args.batch_size
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=args.lr, weight_decay=args.weight_decay)
        self.eval_epoch = args.eval_epoch
        self.task.dataset.set_neg_num(args.neg_num)

    def train(self):
        epoch_iter = tqdm(range(self.max_epoch))
        scores = dict()
        for epoch in epoch_iter:
            self.task.dataset.get_complete_events_features()
            loss = self._full_train_step()
            print("loss: ", loss)

        epoch_test_iter = tqdm(range(self.eval_epoch))
        for epoch in epoch_test_iter:
            score = self._full_test_step()
            scores[epoch] = score.tolist()
        event_label = self.event_label.tolist()
        the_score_all = []
        for i in range(len(event_label)):
            the_score = 0.0
            for j in range(self.eval_epoch):
                the_score += scores[j][i]
            the_score_all.append(the_score)
        ap = self.ap(event_label, the_score_all)
        auc = self.auc(event_label, the_score_all)
        print("AP:", ap)
        print("AUC:", auc)

    def _full_train_step(self):
        self.model.train()
        event_list, neg_event_list, pos_event_list, neg_context_list, neg_entity_list, event_mask, type_num = self.task.dataset.get_batch(
            self.batch_size, device=self.device)
        num_of_batch = len(event_list)
        end_loss = 0.0
        train_epoch_iter = tqdm(range(num_of_batch))
        for i in train_epoch_iter:
            self.optimizer.zero_grad()
            loss = self.model(event_list[i], pos_event_list[i], neg_event_list[i], neg_context_list[i], neg_entity_list[i], event_mask[i], type_num[i])
            loss.backward()
            self.optimizer.step()
            end_loss += loss
        return end_loss / num_of_batch

    def _full_test_step(self):
        self.model.eval()
        event_list, neg_event_list, pos_event_list, neg_context_list, neg_entity_list, event_mask, type_num = self.task.dataset.get_batch(
            self.batch_size, shuffle=False, device=self.device)
        num_of_batch = len(event_list)
        scores = None
        with torch.no_grad():
            test_epoch_iter = tqdm(range(num_of_batch))
            for i in test_epoch_iter:
                self.optimizer.zero_grad()
                score = self.model(event_list[i], pos_event_list[i], neg_event_list[i], neg_context_list[i],
                                   neg_entity_list[i], event_mask[i], type_num[i])
                if scores is None:
                    scores = score
                else:
                    scores = torch.cat((scores, score), 0)

        return scores



from abc import abstractmethod
import dgl
import torch as th
from torch import nn
from tqdm import tqdm
import torch
import torch.nn.functional as F
from openhgnn.trainerflow import BaseFlow, register_flow
from openhgnn.models import build_model
from openhgnn.utils import extract_embed, EarlyStopping, get_nodes_dict, add_reverse_edges
from torch.utils.data import DataLoader
from openhgnn.tasks import build_task
import os
from sklearn import metrics
import numpy as np


@register_flow("DHNE_trainer")
class DHNE_trainer(BaseFlow):
    """
    DHNE trainer flow.
    Supported Model: DHNE
    Supported Datase: drug, GPS, MovieLens, wordnet
    The trainerflow supports hypergraph embedding task.
    For more details, please refer to the original paper: https://arxiv.org/abs/1711.10146
    """
    def __init__(self, args):
        super(DHNE_trainer, self).__init__(args)
        self.hg = None

        self.args = args
        self.logger = self.args.logger
        self.model_name = args.model_name
        self.model = args.model
        self.task = build_task(args)
        self.train_dataset = self.task.train_data
        self.val_dataset = self.task.val_data
        self.device = self.args.device
        self.args.nums_type = self.train_dataset.nums_type
        self.model = build_model(self.model).build_model_from_args(
            self.args).to(self.device)
        self.optimizer = torch.optim.RMSprop(
            self.model.parameters(), lr=args.lr)

        self.train_dataloader = DataLoader(self.train_dataset,
                                           batch_size=self.args.batch_size,
                                           shuffle=True,
                                           num_workers=4,
                                           collate_fn=self.collate_fn)
        self.val_dataloader = DataLoader(self.val_dataset,
                                         batch_size=self.args.batch_size,
                                         shuffle=True,
                                         num_workers=4,
                                         collate_fn=self.collate_fn)

    def collate_fn(self, batches):
        encode1, encode2, encode3 = [], [], []
        labels = []
        for encodes, label in batches:
            encode1.append(torch.from_numpy(encodes[0]))
            encode2.append(torch.from_numpy(encodes[1]))
            encode3.append(torch.from_numpy(encodes[2]))
            labels.append(torch.from_numpy(label))

        return [torch.concat(encode1).type(torch.float32), torch.concat(encode2).type(torch.float32), torch.concat(encode3).type(torch.float32)], torch.concat(labels)

    def preprocess(self):

        pass

    def train(self):
        total_metric = {}
        max_auc = 0
        max_auc_epoch = 0
        max_auc_epoch_loss = 0
        for epoch in tqdm(range(self.max_epoch)):
            loss = self._mini_train_step()
            if epoch % self.evaluate_interval == 0:
                val_metric = self._test_step()
                self.logger.train_info(
                    f"Epoch: {epoch:03d}, train loss: {loss:.4f}. " +
                    "encode_0:{encode_0},encode_1:{encode_1},encode_2:{encode_2},label:{label},auc_micro:{auc_micro}".format(**val_metric))
                total_metric[epoch] = val_metric
                if val_metric['auc_micro'] > max_auc:
                    max_auc = val_metric['auc_micro']
                    max_auc_epoch = epoch
                    max_auc_epoch_loss = loss
        print("[Best Info] the best training results:")
        print(f"Epoch: {max_auc_epoch:03d}, train loss: {max_auc_epoch_loss:.4f}. " +
              "encode_0:{encode_0},encode_1:{encode_1},encode_2:{encode_2},label:{label},auc_micro:{auc_micro}".format(
                  **total_metric[max_auc_epoch]))
        save_path = os.path.dirname(os.path.abspath('__file__'))+'/openhgnn/output/'+self.model_name
        th.save(self.model.state_dict(), save_path+'/DHNE.pth')

    def _full_train_setp(self):
        pass

    def _mini_train_step(self,):
        self.model.train()
        all_loss = 0
        loader_tqdm = tqdm(self.train_dataloader, ncols=120)
        for blocks, target in loader_tqdm:
            blocks = [b.to(self.device) for b in blocks]
            target = target.to(self.device)
            logits = self.model(blocks)
            loss = self.loss_calculation(blocks+[target], logits)
            all_loss += loss.item()
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
        return all_loss

    def sparse_autoencoder_error(self, y_true, y_pred):
        return torch.mean(torch.square(torch.sign(y_true)*(y_true-y_pred)), axis=-1)

    def loss_calculation(self, targets, preds):
        loss = 0
        for i in range(3):
            target = targets[i]
            pred = preds[i]
            loss += self.sparse_autoencoder_error(target, pred)
        label_target = targets[-1]
        label_target = label_target.reshape(-1, 1)
        label_pred = preds[-1]
        loss = loss.mean()
        loss += F.binary_cross_entropy_with_logits(label_pred, label_target)
        return loss

    def _test_step(self):
        self.model.eval()
        metric = {"encode_0": 0, "encode_1": 0, "encode_2": 0, "label": 0}
        with th.no_grad():
            loader_tqdm = tqdm(self.val_dataloader, ncols=120)
            num = 0
            y_true = []
            y_score = []
            for blocks, target in loader_tqdm:
                blocks = [b.to(self.device) for b in blocks]
                targets = target.to(self.device)
                logits = self.model(blocks)
                for i in range(3):
                    pred = logits[i]
                    target = blocks[i]
                    metric["encode_%d" % i] += F.mse_loss(pred, target)
                label_logit = torch.sigmoid(logits[-1])
                label_pred = (label_logit > 0.5).int()
                label_pred = label_pred.reshape(-1)
                metric["label"] += (label_pred == targets).float().mean()
                y_true.append(targets.cpu().numpy())
                y_score.append(label_logit.cpu().numpy())
                num += 1
        metric = {k: v.cpu().numpy() / num for k, v in metric.items()}
        y_true = np.hstack(y_true)
        y_score = np.vstack(y_score).reshape(-1)
        auc_micro = metrics.roc_auc_score(y_true, y_score, average='micro')
        metric['auc_micro'] = auc_micro
        return metric

    def save_checkpoint(self):
        if self._checkpoint and hasattr(self.model, "_parameters()"):
            torch.save(self.model.state_dict(), self._checkpoint)



import copy

import dgl
import numpy as np
import torch as th
from sklearn.metrics import f1_score
from torch import nn
from tqdm import tqdm
import torch
import torch.nn.functional as F
from . import BaseFlow, register_flow
from ..models import build_model
from ..models.LogReg import LogReg

@register_flow("SHGP_trainer")
class SHGPTrainer(BaseFlow):
    """
    Pre-Training of GNNs on Heterogeneous Graphs
    """

    def __init__(self, args):
        super(SHGPTrainer, self).__init__(args)
        self.args = args
        self.target_type = args.target_type
        self.logger = args.logger
        self.train_percent = args.train_percent
        self.dataset = self.task.dataset
        self.label = self.dataset.label
        self.ft_dict = self.dataset.ft_dict
        self.adj_dict = self.dataset.adj_dict
        self.target_type = args.target_type
        self.num_cluster = int(
            self.ft_dict[self.target_type].shape[0] * args.compress_ratio)
        self.num_class = np.unique(self.label[self.target_type][0]).shape[0]
        self.layer_shape = []
        self.input_layer_shape = dict([(k, self.ft_dict[k].shape[1]) for k in self.ft_dict.keys()])
        self.hidden_layer_shape = [dict.fromkeys(self.ft_dict.keys(), l_hid) for l_hid in args.hidden_dim]
        self.output_layer_shape = dict.fromkeys(self.ft_dict.keys(), self.num_cluster)

        self.layer_shape.append(self.input_layer_shape)
        self.layer_shape.extend(self.hidden_layer_shape)
        self.layer_shape.append(self.output_layer_shape)

        self.net_schema = dict([(k, list(self.adj_dict[k].keys())) for k in self.adj_dict.keys()])
        self.args.net_schema = self.net_schema
        self.args.layer_shape = self.layer_shape
        self.args.label = self.label
        self.model = build_model(self.model).build_model_from_args(self.args, self.hg).to(self.device)
        # self.model = ATT_HGCN(
        #     net_schema=self.net_schema,
        #     layer_shape=self.layer_shape,
        #     label_keys=list(self.label.keys()),
        #     type_fusion=args.type_fusion,
        #     type_att_size=args.type_att_size,
        # )

    def preprocess(self):
        # super(SHGPTrainer, self).preprocess()
        if self.args.cuda and torch.cuda.is_available():
            self.model.cuda()
            for k in self.ft_dict:
                self.ft_dict[k] = self.ft_dict[k].cuda()
            for k in self.adj_dict:
                for kk in self.adj_dict[k]:
                    self.adj_dict[k][kk] = self.adj_dict[k][kk].cuda()
            for k in self.label:
                for i in range(len(self.label[k])):
                    self.label[k][i] = self.label[k][i].cuda()
        return

    def train(self):
        self.preprocess()
        if not self.args.load_from_pretrained:
            best = 1e9
            loss_list = []
            init_pseudo_label = 0
            pseudo_pseudo_label = 0
            optimizer = torch.optim.Adam(self.model.parameters(), lr=self.args.lr, weight_decay=self.args.l2_coef)
            for epoch in range(self.args.epochs):
                self.model.train()
                optimizer.zero_grad()
                logits, embd, attention_dict = self.model(self.ft_dict, self.adj_dict)
                target_embd = embd[self.target_type]

                if epoch == 0:
                    init_pseudo_label = init_lpa(self.adj_dict, self.ft_dict, self.target_type, self.num_cluster)
                    pseudo_label_dict = init_pseudo_label
                elif epoch < self.args.warm_epochs:
                    pseudo_label_dict = init_pseudo_label
                else:
                    pseudo_label_dict = att_lpa(self.adj_dict, init_pseudo_label, attention_dict, self.target_type, self.num_cluster)
                    init_pseudo_label = pseudo_label_dict
                label_predict = torch.argmax(pseudo_label_dict[self.target_type], dim=1)
                logits = F.log_softmax(logits[self.target_type], dim=1)
                loss_train = F.nll_loss(logits, label_predict.long().detach())
                loss_train.backward()
                optimizer.step()
                loss_list.append(loss_train)
                if loss_train < best:
                    best = loss_train

                print(
                    'epoch: {:3d}'.format(epoch),
                    'train loss: {:.4f}'.format(loss_train.item()),
                )
            torch.save(self.model.state_dict(), "openhgnn/dataset/data/test/test")
        else:
            self.model.load_state_dict(torch.load("openhgnn/dataset/data/test/test"))

        # evaluate
        logits, embd, _ = self.model(self.ft_dict, self.adj_dict)
        target_embd = embd[self.target_type]
        label_target = self.label[self.target_type]
        true_label = label_target[0]
        idx_train = label_target[1]
        idx_val = label_target[2]
        idx_test = label_target[3]
        self.evaluate(target_embd, idx_train, idx_val, idx_test, true_label, self.num_class, isTest=True)


    def evaluate(self, embeds, idx_train, idx_val, idx_test, labels, num_class, isTest=True):
        print("----------------------------start evaluating----------------------------------")
        hid_units = embeds.shape[1]
        nb_classes = num_class
        xent = nn.CrossEntropyLoss()
        train_embs = embeds[idx_train]
        val_embs = embeds[idx_val]
        test_embs = embeds[idx_test]
        train_lbls = labels[idx_train]
        val_lbls = labels[idx_val]
        test_lbls = labels[idx_test]

        run_num = 10
        epoch_num = 50
        accs = []
        micro_f1s = []
        macro_f1s = []
        macro_f1s_val = []
        micro_f1s_val = []
        for _ in range(run_num):
            log = LogReg(hid_units, nb_classes)
            # log.load_state_dict(load_LogReg(torch.load(args.pretrain_model_dir)), strict=False)
            opt = torch.optim.Adam(log.parameters(), lr=0.01, weight_decay=0.0)
            if torch.cuda.is_available():
                log.cuda()

            val_accs = [];
            test_accs = []
            val_micro_f1s = [];
            test_micro_f1s = []
            val_macro_f1s = [];
            test_macro_f1s = []

            for iter_ in range(epoch_num):
                # train
                log.train()
                opt.zero_grad()

                logits = log(train_embs)
                loss = xent(logits, train_lbls)

                loss.backward(retain_graph=True)
                opt.step()

                # val
                logits = log(val_embs)
                preds = torch.argmax(logits, dim=1)

                val_acc = torch.sum(preds == val_lbls).float() / val_lbls.shape[0]
                if torch.cuda.is_available():
                    val_f1_macro = f1_score(val_lbls.cpu(), preds.cpu(), average='macro')
                    val_f1_micro = f1_score(val_lbls.cpu(), preds.cpu(), average='micro')
                else:
                    val_f1_macro = f1_score(val_lbls.cpu(), preds, average='macro')
                    val_f1_micro = f1_score(val_lbls.cpu(), preds, average='micro')

                val_accs.append(val_acc.item())
                val_macro_f1s.append(val_f1_macro)
                val_micro_f1s.append(val_f1_micro)

                # test
                logits = log(test_embs)
                preds = torch.argmax(logits, dim=1)

                test_acc = torch.sum(preds == test_lbls).float() / test_lbls.shape[0]
                if torch.cuda.is_available():
                    test_f1_macro = f1_score(test_lbls.cpu(), preds.cpu(), average='macro')
                    test_f1_micro = f1_score(test_lbls.cpu(), preds.cpu(), average='micro')
                else:
                    test_f1_macro = f1_score(test_lbls.cpu(), preds, average='macro')
                    test_f1_micro = f1_score(test_lbls.cpu(), preds, average='micro')

                test_accs.append(test_acc.item())
                test_macro_f1s.append(test_f1_macro)
                test_micro_f1s.append(test_f1_micro)

            max_iter = val_accs.index(max(val_accs))
            accs.append(test_accs[max_iter])

            max_iter = val_macro_f1s.index(max(val_macro_f1s))
            macro_f1s.append(test_macro_f1s[max_iter])
            macro_f1s_val.append(val_macro_f1s[max_iter])  ###

            max_iter = val_micro_f1s.index(max(val_micro_f1s))
            micro_f1s.append(test_micro_f1s[max_iter])
            micro_f1s_val.append(val_micro_f1s[max_iter])

        if isTest:
            print("\t[Classification-test] Macro-F1: {:.4f} ({:.4f}) | Micro-F1: {:.4f} ({:.4f})".format(
                np.mean(macro_f1s), np.std(macro_f1s), np.mean(micro_f1s), np.std(micro_f1s)))


def gen_rand_label(ft_dict,num_cluster):
    rand_label_dict = dict.fromkeys(ft_dict.keys())
    for k in ft_dict:
        rand_label = torch.randint(num_cluster, (ft_dict[k].shape[0],))
        rand_label = F.one_hot(rand_label, num_cluster).type(dtype=torch.float32)
        rand_label_dict[k] = rand_label

    return rand_label_dict



def cat_homo_adj(adj_dict):
    for k in adj_dict:
        print(k)

    return None



def lpa(init_label_dict,adj_dict,target_type,num_cluster,max_iter=1000):
    pseudo_label_dict = copy.deepcopy(init_label_dict)
    for k in pseudo_label_dict:
        if torch.cuda.is_available():
            pseudo_label_dict[k] = pseudo_label_dict[k].cuda()
    net_schema = dict([(k, list(adj_dict[k].keys())) for k in adj_dict.keys()])
    target_label_list=[]
    soft_label=0
    for i in range(max_iter):
        for k in net_schema:
            k_nbs_label_list = []
            for kk in net_schema[k]:
                try:
                    soft_label = torch.spmm(adj_dict[k][kk], pseudo_label_dict[kk])
                except KeyError as ke:
                    soft_label = torch.spmm(adj_dict[k][kk], pseudo_label_dict[k])
                finally:
                    k_nbs_label_list.append(soft_label)
            new_k_label = torch.cat([nb_label.unsqueeze(1) for nb_label in k_nbs_label_list], 1)

            new_k_label = new_k_label.sum(1)
            new_k_label = torch.argmax(new_k_label, dim=1)
            new_k_label = F.one_hot(new_k_label, num_cluster).type(dtype=torch.float32)
            pseudo_label_dict[k] = new_k_label
            if k==target_type:
                target_label_list.append(new_k_label)
        if len(target_label_list)>1:
            if target_label_list[-2].equal(target_label_list[-1]):
                break

    return pseudo_label_dict


def init_lpa(adj_dict, ft_dict, target_type, num_cluster):
    run_num = 1
    for i in range(run_num):
        init_label_dict = gen_rand_label(ft_dict, num_cluster)
        pseudo_label_dict = lpa(init_label_dict, adj_dict, target_type, num_cluster)

    return pseudo_label_dict


def att_lpa(adj_dict, init_pseudo_label, attention_dict, target_type, num_cluster, max_iter=1000):
    pseudo_label_dict = copy.deepcopy(init_pseudo_label)
    current_label_dict = copy.deepcopy(init_pseudo_label)

    for k in pseudo_label_dict:
        if torch.cuda.is_available():
            pseudo_label_dict[k] = pseudo_label_dict[k].cuda()
    for k in current_label_dict:
        if torch.cuda.is_available():
            current_label_dict[k] = current_label_dict[k].cuda()
    net_schema = dict([(k, list(adj_dict[k].keys())) for k in adj_dict.keys()])
    target_label_list = []
    soft_label = 0
    for _ in range(max_iter):
        for m in range(len(attention_dict)):
            for k in net_schema:
                k_nbs_label_list = []
                k_nbs_label_list.append(pseudo_label_dict[k])
                for kk in net_schema[k]:
                    try:
                        soft_label = torch.spmm(adj_dict[k][kk], current_label_dict[kk])
                    except KeyError as ke:
                        soft_label = torch.spmm(adj_dict[k][kk], current_label_dict[k])
                    finally:
                        k_nbs_label_list.append(soft_label)
                pseudo_label_dict[k] = torch.cat([nb_label.unsqueeze(1) for nb_label in k_nbs_label_list], 1).mul(
                    attention_dict[m][k].unsqueeze(-1)).sum(1)

            for k in net_schema:
                current_label_dict[k] = pseudo_label_dict[k]

        for k in net_schema:
            new_k_label = torch.argmax(pseudo_label_dict[k], dim=1)
            if k == target_type:
                target_label_list.append(new_k_label)
            pseudo_label_dict[k] = F.one_hot(new_k_label, num_cluster).type(dtype=torch.float32)
            current_label_dict[k] = pseudo_label_dict[k]
        if len(target_label_list) > 1:
            if target_label_list[-2].equal(target_label_list[-1]):
                break

    return pseudo_label_dict









from . import BaseTask, register_task
from ..dataset import build_dataset


@register_task("embedding")
class Embedding(BaseTask):
    """Embedding task. Generate embeddings for nodes in the graph."""

    def __init__(self, args):
        super(Embedding, self).__init__()
        self.n_dataset = args.dataset
        self.dataset = build_dataset(args.dataset, 'embedding')

    def get_graph(self):
        return getattr(self.dataset, 'g', self.dataset[0])

    def get_loss_fn(self):
        pass

    def evaluate(self):
        pass



"""Dataset adapters for re-purposing a dataset for a different kind of training task."""

import os
import json
import numpy as np
from dgl.data import utils, DGLDataset
from dgl import backend as F
import dgl
from dgl.dataloading.negative_sampler import GlobalUniform, PerSourceUniform
import torch as th
from dgl import DGLHeteroGraph
import torch

__all__ = ['AsNodeClassificationDataset', 'AsLinkPredictionDataset']


class AsNodeClassificationDataset(DGLDataset):
    """Repurpose a dataset for a standard semi-supervised transductive
    node prediction task.

    The class converts a given dataset into a new dataset object that:

      - Contains only one heterogeneous graph, accessible from ``dataset[0]``.
      - The graph stores:

        - Node labels in ``g.nodes[target_ntype].data['label']``.
        - Train/val/test masks in ``g.nodes[target_ntype].data['train_mask']``, ``g.nodes[target_ntype].data['val_mask']``,
          and ``g.nodes[target_ntype].data['test_mask']`` respectively.

      - In addition, the dataset contains the following attributes:

        - ``num_classes``, the number of classes to predict.
        - ``train_idx``, ``val_idx``, ``test_idx``, train/val/test indexes.

    The class will keep only the first graph in the provided dataset and
    generate train/val/test masks according to the given spplit ratio. The generated
    masks will be cached to disk for fast re-loading. If the provided split ratio
    differs from the cached one, it will re-process the dataset properly.

    Parameters
    ----------
    data : DGLDataset or DGLHeteroGraph
        The dataset or graph to be converted.
    name : str
        The dataset name. Optional when data is DGLDataset. Required when data is DGLHeteroGraph.
    labeled_nodes_split_ratio : (float, float, float), optional
        Split ratios for training, validation and test sets. Must sum to 1. If None, we will use the train_mask,
        val_mask and test_mask from the original graph.
    prediction_ratio : float, optional
        The ratio of number of prediction nodes to all unlabeled nodes. Prediction_ratio ranges from 0 to 1.
        If None, we will use the pred_mask from the original graph.
    target_ntype : str
        The node type to add split mask for.
    label_feat_name: str, optional
        The feature name of label.
        If None, we will use the name "label".
    label_mask_feat_name: str, optional
        The feature name of the mask indicating the indices of nodes with labels. None means that all nodes are labeled.

    Attributes
    ----------
    num_classes : int
        Number of classes to predict.
    train_idx : Tensor
        An 1-D integer tensor of training node IDs.
    val_idx : Tensor
        An 1-D integer tensor of validation node IDs.
    test_idx : Tensor
        An 1-D integer tensor of test node IDs.
    pred_idx : Tensor
        An 1-D integer tensor of prediction node IDs.
    """

    def __init__(self,
                 data,
                 name=None,
                 labeled_nodes_split_ratio=None,
                 prediction_ratio=None,
                 target_ntype=None,
                 label_feat_name='label',
                 label_mask_feat_name=None,
                 **kwargs):

        self.label_feat_name = label_feat_name
        self.prediction_ratio = prediction_ratio
        self.label_mask_feat_name = label_mask_feat_name
        if isinstance(data, DGLDataset):
            self.dataset = data
            self.g = data[0]
            if name is None:
                name = self.dataset.name
        elif isinstance(data, DGLHeteroGraph):
            self.dataset = None
            self.g = data
            assert name is not None, \
                "Name is required when data is a graph."
        else:
            raise ValueError("Invalid data type.")

        self.split_ratio = kwargs.pop('split_ratio', None)  # for compatibility
        if labeled_nodes_split_ratio is not None:
            self.split_ratio = labeled_nodes_split_ratio

        self.target_ntype = target_ntype
        super().__init__(name + '-as-nodepred',
                         hash_key=(self.split_ratio, target_ntype, name, 'nodepred'), **kwargs)

    def process(self):
        if self.label_feat_name not in self.g.nodes[self.target_ntype].data:
            raise ValueError("Missing node labels. Make sure labels are stored "
                             "under name {}.".format(self.label_feat_name))

        if self.split_ratio is None:
            if self.verbose:
                print('Split ratio is not provided, '
                      'we will use the train_mask, val_mask and test_mask from the original graph.')
        else:
            if self.verbose:
                print('Generating train/val/test masks...')
            # utils.add_nodepred_split(self, self.split_ratio, self.target_ntype)
            self.gene_mask(self.split_ratio, self.target_ntype, )
        if self.prediction_ratio is None:
            if self.verbose:
                print("Prediction ratio is not provided, we will use the pred_mask from the original graph.")
        elif self.label_mask_feat_name is not None:
            self.gene_pred_mask(self.prediction_ratio, self.target_ntype)
        else:
            if self.verbose:
                print('All nodes have label, will not predict.')
        self._set_split_index(self.target_ntype)
        self.multi_label = getattr(self.dataset, 'multi_label', None)
        if self.multi_label is None:
            self.multi_label = len(self.g.nodes[self.target_ntype].data[self.label_feat_name].shape) == 2

        self.num_classes = getattr(self.dataset, 'num_classes', None)
        if self.num_classes is None:
            if self.multi_label:
                self.num_classes = self.g.nodes[self.target_ntype].data[self.label_feat_name].shape[1]
            else:
                self.num_classes = len(F.unique(self.g.nodes[self.target_ntype].data[self.label_feat_name]))

        self.meta_paths = getattr(self.dataset, 'meta_paths', None)
        self.meta_paths_dict = getattr(self.dataset, 'meta_paths_dict', None)

    def gene_pred_mask(self, ratio, ntype):
        idx_tensor = torch.where(self.g.nodes[ntype].data[self.label_mask_feat_name] == 0)[0]
        idx = idx_tensor.tolist()
        len_nodes = len(self.g.nodes[ntype].data[self.label_mask_feat_name])
        n = len(idx)
        np.random.shuffle(idx)
        n_pred = int(n * ratio)
        pred_mask = utils.generate_mask_tensor(utils.idx2mask(idx[:n_pred], len_nodes))
        self.g.nodes[ntype].data['pred_mask'] = pred_mask

    def gene_mask(self, ratio, ntype):
        if len(ratio) != 3:
            raise ValueError(f'Split ratio must be a float triplet but got {ratio}.')
        if self.label_mask_feat_name is None:
            idx_tensor = self.g.nodes(ntype)
        else:
            idx_tensor = torch.nonzero(self.g.nodes[ntype].data[self.label_mask_feat_name]).squeeze(1)
        idx = idx_tensor.tolist()
        len_nodes = len(self.g.nodes(ntype))
        n = len(idx)
        # idx = np.arange(0, n)
        np.random.shuffle(idx)
        n_train, n_val, n_test = int(n * ratio[0]), int(n * ratio[1]), int(n * ratio[2])
        train_mask = utils.generate_mask_tensor(utils.idx2mask(idx[:n_train], len_nodes))
        val_mask = utils.generate_mask_tensor(utils.idx2mask(idx[n_train:n_train + n_val], len_nodes))
        test_mask = utils.generate_mask_tensor(utils.idx2mask(idx[n_train + n_val:], len_nodes))
        self.g.nodes[ntype].data['train_mask'] = train_mask
        self.g.nodes[ntype].data['val_mask'] = val_mask
        self.g.nodes[ntype].data['test_mask'] = test_mask

    def has_cache(self):
        return os.path.isfile(os.path.join(self.save_path, 'graph_{}.bin'.format(self.hash)))

    def load(self):
        with open(os.path.join(self.save_path, 'info_{}.json'.format(self.hash)), 'r') as f:
            info = json.load(f)
            if (info['split_ratio'] != self.split_ratio
                    or info['target_ntype'] != self.target_ntype):
                raise ValueError('Provided split ratio is different from the cached file. '
                                 'Re-process the dataset.')
            self.split_ratio = info['split_ratio']
            self.target_ntype = info['target_ntype']
            self.num_classes = info['num_classes']
            self.meta_paths_dict = info['meta_paths_dict']
            self.meta_paths = info['meta_paths']
            self.multi_label = info['multi_label']
            self.label_feat_name = info['label_feat_name']
            self.prediction_ratio = info['prediction_ratio']
            self.label_mask_feat_name = info['label_mask_feat_name']
        gs, _ = utils.load_graphs(os.path.join(self.save_path, 'graph_{}.bin'.format(self.hash)))
        self.g = gs[0]
        self._set_split_index(self.target_ntype)

    def save(self):
        utils.save_graphs(os.path.join(self.save_path, 'graph_{}.bin'.format(self.hash)), [self.g])
        with open(os.path.join(self.save_path, 'info_{}.json'.format(self.hash)), 'w') as f:
            json.dump({
                'split_ratio': self.split_ratio,
                'target_ntype': self.target_ntype,
                'num_classes': self.num_classes,
                'multi_label': self.multi_label,
                'meta_paths_dict': self.meta_paths_dict,
                'meta_paths': self.meta_paths,
                'label_feat_name': self.label_feat_name,
                'prediction_ratio': self.prediction_ratio,
                'label_mask_feat_name': self.label_mask_feat_name}, f)

    def __getitem__(self, idx):
        return self.g

    def __len__(self):
        return 1

    def _set_split_index(self, ntype):
        """Add train_idx/val_idx/test_idx as dataset attributes according to corresponding mask."""
        ndata = self.g.nodes[self.target_ntype].data
        self.train_idx = None
        self.val_idx = None
        self.test_idx = None
        self.pred_idx = None
        if 'train_mask' in ndata:
            self.train_idx = F.nonzero_1d(ndata['train_mask'])
        if 'val_mask' in ndata:
            self.val_idx = F.nonzero_1d(ndata['val_mask'])
        if 'test_mask' in ndata:
            self.test_idx = F.nonzero_1d(ndata['test_mask'])
        if 'pred_mask' in ndata:
            self.pred_idx = F.nonzero_1d(ndata['pred_mask'])
        elif self.label_mask_feat_name is not None:
            if self.verbose:
                print('No prediction mask exists, will predict all missing labels.')
            idx_tensor = torch.where(self.g.nodes[ntype].data[self.label_mask_feat_name] == 0)[0]
            self.pred_idx = idx_tensor

    def get_split(self, *args, **kwargs):
        return self.train_idx, self.val_idx, self.test_idx

    def get_labels(self):
        return self.g.nodes[self.target_ntype].data[self.label_feat_name]

    @property
    def category(self):
        return self.target_ntype


class AsLinkPredictionDataset(DGLDataset):
    """Repurpose a dataset for link prediction task.

    The created dataset will include data needed for link prediction.
    It will keep only the first graph in the provided dataset and
    generate train/val/test edges according to the given split ratio,
    and the correspondent negative edges based on the neg_ratio. The generated
    edges will be cached to disk for fast re-loading. If the provided split ratio
    differs from the cached one, it will re-process the dataset properly.

    Parameters
    ----------
    dataset : DGLDataset
        The dataset to be converted.
    split_ratio : (float, float, float), optional
        Split ratios for training, validation and test sets. Must sum to one.
    neg_ratio : int, optional
        Indicate how much negative samples to be sampled
        The number of the negative samples will be equal or less than neg_ratio * num_positive_edges.
    target_link : list[tuple[str, str, str]]
        The edge types on which predictions are make.
    target_link_r : list[tuple[str, str, str]], optional
        The reverse edge types of the target links. Used to remove reverse edges of val/test edges from train graph.
    neg_sampler : str, optional
        Indicate how negative edges of val/test edges are sampled. 'global' or 'per_source'.

    Attributes
    -------
    train_graph: DGLHeteroGraph
        The DGLHeteroGraph for training
    pos_val_graph: DGLHeteroGraph
        The DGLHeteroGraph containing positive validation edges
    pos_test_graph: DGLHeteroGraph
        The DGLHeteroGraph containing positive test edges
    neg_val_graph: DGLHeteroGraph
        The DGLHeteroGraph containing negative validation edges
    neg_test_graph: DGLHeteroGraph
        The DGLHeteroGraph containing negative test edges
    """

    def __init__(self,
                 dataset,
                 target_link,
                 target_link_r,
                 split_ratio=None,
                 neg_ratio=3,
                 neg_sampler='global',
                 **kwargs):
        self.g = dataset[0]
        self.num_nodes = self.g.num_nodes()
        self.dataset = dataset
        self.split_ratio = split_ratio
        self.target_link = target_link
        self.target_link_r = target_link_r
        self.neg_ratio = neg_ratio
        self.neg_sampler = neg_sampler
        super().__init__(dataset.name + '-as-linkpred', hash_key=(
            neg_ratio, target_link, target_link_r, split_ratio, neg_sampler, dataset.name, 'linkpred'), **kwargs)

    def process(self):
        if self.split_ratio is None:
            for etype in self.target_link:
                for mask in ['train_mask', 'val_mask', 'test_mask']:
                    assert mask in self.g.edges[etype].data, \
                        "{} is not provided for edge type {}, please specify split_ratio to generate the masks".format(
                            mask, etype)

        else:
            ratio = self.split_ratio
            for etype in self.target_link:
                n = self.g.num_edges(etype)
                n_train, n_val, n_test = int(n * ratio[0]), int(n * ratio[1]), int(n * ratio[2])
                idx = np.random.permutation(n)
                train_idx = idx[:n_train]
                val_idx = idx[n_train:n_train + n_val]
                test_idx = idx[n_train + n_val:]
                train_mask = th.zeros(n).bool()
                train_mask[train_idx] = True
                val_mask = th.zeros(n).bool()
                val_mask[val_idx] = True
                test_mask = th.zeros(n).bool()
                test_mask[test_idx] = True
                self.g.edges[etype].data['train_mask'] = train_mask
                self.g.edges[etype].data['val_mask'] = val_mask
                self.g.edges[etype].data['test_mask'] = test_mask

        # create val and test graph(pos and neg respectively)
        self.pos_val_graph, self.neg_val_graph = self._get_pos_and_neg_graph('val')
        self.pos_test_graph, self.neg_test_graph = self._get_pos_and_neg_graph('test')

        self.pred_edges = getattr(self.dataset, 'pred_edges', None)
        if self.pred_edges is not None:
            self.pred_graph = dgl.heterograph(self.pred_edges,
                                              {ntype: self.g.num_nodes(ntype) for ntype in self.g.ntypes})

        # create train graph
        train_graph = self.g
        for i, etype in enumerate(self.target_link):
            # remove val and test edges
            train_graph = dgl.remove_edges(train_graph,
                                           th.cat((self.pos_val_graph.edges[etype].data[dgl.EID],
                                                   self.pos_test_graph.edges[etype].data[dgl.EID])),
                                           etype)
            # remove reverse edges of val and test edges
            if self.target_link_r is not None:
                reverse_etype = self.target_link_r[i]
                train_graph = dgl.remove_edges(train_graph, th.arange(train_graph.num_edges(reverse_etype)),
                                               reverse_etype)
                edges = train_graph.edges(etype=etype)
                train_graph = dgl.add_edges(train_graph, edges[1], edges[0], etype=reverse_etype)
        self.train_graph = train_graph

        self.meta_paths = getattr(self.dataset, 'meta_paths', None)
        self.meta_paths_dict = getattr(self.dataset, 'meta_paths_dict', None)

    def _get_pos_and_neg_graph(self, split):
        if self.neg_sampler == 'global':
            neg_sampler = GlobalUniform(self.neg_ratio)
        elif self.neg_sampler == 'per_source':
            neg_sampler = PerSourceUniform(self.neg_ratio)
        else:
            raise ValueError('Unsupported neg_sampler')
        edges = {
            etype: th.nonzero(self.g.edges[etype].data['{}_mask'.format(split)]).squeeze()
            for etype in self.target_link}
        pos_graph = dgl.edge_subgraph(self.g, edges, relabel_nodes=False, store_ids=True)
        neg_edges = getattr(self.dataset, 'neg_{}_edges'.format(split), neg_sampler(self.g, edges))
        neg_graph = dgl.heterograph(neg_edges, {ntype: pos_graph.num_nodes(ntype) for ntype in pos_graph.ntypes})
        return pos_graph, neg_graph

    def has_cache(self):
        return os.path.isfile(os.path.join(self.save_path, 'graph_{}.bin'.format(self.hash)))

    def load(self):
        gs, _ = utils.load_graphs(
            os.path.join(self.save_path, 'graph_{}.bin'.format(self.hash)))

        self.train_graph, self.pos_val_graph, self.pos_test_graph, self.neg_val_graph, self.neg_test_graph = \
            gs[0], gs[1], gs[2], gs[3], gs[4]

        with open(os.path.join(self.save_path, 'info_{}.json'.format(self.hash)), 'r') as f:
            info = json.load(f)
            self.split_ratio = info["split_ratio"]
            self.neg_ratio = info["neg_ratio"]
            self.target_link = info["target_link"]
            self.target_link_r = info["target_link_r"]
            self.neg_sampler = info["neg_sampler"]
            self.meta_paths_dict = info["meta_paths_dict"]
            self.meta_paths = info["meta_paths"]

    def save(self):
        utils.save_graphs(os.path.join(self.save_path, 'graph_{}.bin'.format(self.hash)),
                          [self.train_graph, self.pos_val_graph, self.pos_test_graph, self.neg_val_graph,
                           self.neg_test_graph])
        with open(os.path.join(self.save_path, 'info_{}.json'.format(self.hash)), 'w') as f:
            json.dump({
                'split_ratio': self.split_ratio,
                'neg_ratio': self.neg_ratio,
                'target_link': self.target_link,
                'target_link_r': self.target_link_r,
                'neg_sampler': self.neg_sampler,
                'meta_paths_dict': self.meta_paths_dict,
                'meta_paths': self.meta_paths,
            }, f)

    def get_split(self, *args, **kwargs):
        return self.train_graph, self.pos_val_graph, self.pos_test_graph, self.neg_val_graph, self.neg_test_graph

    def __getitem__(self, idx):
        return self.g

    def __len__(self):
        return 1
